import json
import uuid
from dataclasses import fields
from typing import Any, Dict, List, Optional, Sequence, Union
from uuid import UUID

# Use LangChain features
from langchain_core.agents import AgentAction, AgentFinish
from langchain_core.callbacks import BaseCallbackHandler
from langchain_core.documents import Document
from langchain_core.messages import BaseMessage, ToolMessage
from langchain_core.outputs import LLMResult

from ...expiring_key_value_store import ExpiringKeyValueStore
from ...logger import (
    GenerationConfig,
    Logger,
    RetrievalConfig,
    SpanConfig,
    ToolCallConfig,
    ToolCallError,
)
from ...scribe import scribe
from ..logger import Generation
from ..models import Container, Metadata, SpanContainer, TraceContainer
from .utils import (
    parse_langchain_llm_error,
    parse_langchain_llm_result,
    parse_langchain_messages,
    parse_langchain_model_and_provider,
    parse_langchain_model_parameters,
    parse_langchain_provider,
)

# 20 minutes
DEFAULT_TIMEOUT = 60 * 20


class MaximLangchainTracer(BaseCallbackHandler):
    """
    A callback handler that logs langchain outputs to Maxim logger

    Args:
        logger: Logger: Maxim Logger instance to log outputs
    """

    def __init__(
        self,
        logger: Logger,
        metadata: Optional[Dict[str, Any]] = None,
        eval_config: Optional[Dict[str, List[str]]] = None,
    ) -> None:
        """Initializes the Langchain Tracer
        Args:
            logger: Logger: Maxim Logger instance to log outputs
        """
        super().__init__()
        self.run_inline = True
        self.logger = logger
        self.containers: Dict[str, Container] = {}
        self.metadata_store = ExpiringKeyValueStore()
        self.to_be_evaluated_container_store = ExpiringKeyValueStore()
        self.metadata = None
        self.eval_config = eval_config
        if metadata is not None:
            self.__validate_maxim_metadata(metadata)
            self.metadata = metadata

    def _parse_metadata(self, metadata: Optional[Dict[str, Any]]):
        """
        Parses the metadata
        """
        try:
            if (
                metadata is None
                or not isinstance(metadata, dict)
                or "maxim" not in metadata
            ):
                return
            maxim_metadata = metadata.get("maxim", None)
            if maxim_metadata is None:
                return
            self.__validate_maxim_metadata(maxim_metadata)
            self.metadata = maxim_metadata
        except Exception as e:
            import traceback

            scribe().error(
                "[MaximSDK] Failed to parse metadata: %s\n%s",
                e,
                traceback.format_exc(),
            )

    def __validate_maxim_metadata(self, metadata: Optional[Dict[str, Any]]):
        """
        Validates the metadata
        """
        if metadata is None:
            return
        id_keys = ["session_id", "trace_id", "span_id"]
        present_keys = [key for key in id_keys if key in metadata]
        if len(present_keys) > 1:
            raise ValueError(
                f"Multiple keys found in metadata: {present_keys}. You can pass only one of these."
            )
        valid_keys = [field.name for field in fields(Metadata)]
        invalid_keys = [key for key in metadata if key not in valid_keys]
        if len(invalid_keys) > 0:
            raise ValueError(
                f"Invalid keys found in metadata: {invalid_keys}. Valid keys are {valid_keys}"
            )

    def __get_container_from_metadata(self, run_id: UUID) -> Container:
        """
        Gets the container from the metadata
        """
        maxim_metadata = self.__get_metadata()
        container: Optional[Container] = None
        if maxim_metadata is not None:
            span_id: Optional[str] = maxim_metadata.span_id
            if span_id is not None:
                span_name: Optional[str] = maxim_metadata.span_name
                container = SpanContainer(
                    span_id=span_id,
                    logger=self.logger,
                    span_name=span_name,
                    mark_created=True,
                )
                # we intentionally add trace tags to the span container
                if maxim_metadata.trace_tags is not None:
                    scribe().info(
                        f"[MaximSDK] Adding trace tags to span container: {maxim_metadata.trace_tags}"
                    )
                    container.add_tags(maxim_metadata.trace_tags)
            trace_id = maxim_metadata.trace_id
            if trace_id is not None:
                trace_name = maxim_metadata.trace_name
                container = TraceContainer(
                    trace_id=trace_id,
                    logger=self.logger,
                    trace_name=trace_name,
                    mark_created=True,
                )
                if maxim_metadata.trace_tags is not None:
                    scribe().info(
                        f"[MaximSDK] Adding trace tags to trace container: {maxim_metadata.trace_tags}"
                    )
                    container.add_tags(maxim_metadata.trace_tags)

        if container is None:
            # We will check if the container is created with run_id
            container = self.containers.get(str(run_id))

        if container is None:
            session_id: Optional[str] = (
                maxim_metadata.session_id if maxim_metadata else None
            )
            container = TraceContainer(
                trace_id=str(uuid.uuid4()),
                logger=self.logger,
                trace_name="Trace",
                parent=session_id,
            )
            container.create()
            if maxim_metadata is not None and maxim_metadata.trace_tags is not None:
                container.add_tags(maxim_metadata.trace_tags)

        return container

    def __get_container(
        self, run_id: UUID, parent_run_id: Optional[UUID] = None
    ) -> Optional[Container]:
        """
        Gets the container for the given run_id
        """
        container: Optional[Container] = None
        if parent_run_id is None:
            # This is the first activity in this run
            container = self.__get_container_from_metadata(run_id)
            # This is redundant but just to be safe
            if container is not None:
                self.containers[str(run_id)] = container
        else:
            container = self.containers.get(str(parent_run_id))
            if container is None:
                # Create a trace container and a span in it
                # This is a fallback scenarios where the callbacks are called in random order
                # This ideally should not happen
                scribe().warning(
                    f"[MaximSDK] Couldn't find a container for parent run id {parent_run_id}. Creating a new trace."
                )
                trace_container = TraceContainer(
                    trace_id=str(parent_run_id),
                    logger=self.logger,
                    trace_name="Trace",
                    mark_created=True,
                )
                trace_container.create()
                self.containers[str(parent_run_id)] = trace_container
        return container

    def __get_metadata(
        self, metadata: Optional[Dict[str, Any]] = None
    ) -> Optional[Metadata]:
        """
        Gets the metadata
        """
        if (
            metadata is not None
            and "maxim" in metadata
            and metadata.get("maxim", None) is not None
        ):
            return Metadata(metadata.get("maxim", None))
        if self.metadata is not None:
            return Metadata(self.metadata)
        return None

    def on_llm_start(
        self,
        serialized: dict[str, Any],
        prompts: list[str],
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        tags: Optional[list[str]] = None,
        metadata: Optional[dict[str, Any]] = None,
        **kwargs: Any,
    ) -> Any:
        """Runs when LLM starts"""
        try:
            self._parse_metadata(metadata)
            scribe().debug("[MaximSDK: Langchain] on_llm_start called")
            model, model_parameters = parse_langchain_model_parameters(**kwargs)
            provider = parse_langchain_provider(serialized)
            model, provider = parse_langchain_model_and_provider(model, provider)
            maxim_messages = parse_langchain_messages(prompts)
            last_input_message = ""
            if maxim_messages is not None:
                for message in maxim_messages:
                    if "role" in message and message["role"] == "user":
                        last_input_message = message["content"]
            generation_id = str(run_id)
            maxim_metadata = self.__get_metadata(metadata)
            # Prepare generation config
            generation_config = GenerationConfig(
                id=generation_id,
                name=maxim_metadata.generation_name if maxim_metadata else None,
                provider=provider,
                model=model,
                messages=maxim_messages,
                model_parameters=model_parameters,
                tags=maxim_metadata.generation_tags if maxim_metadata else None,
            )
            # Adding it back to the container
            container = self.__get_container(run_id, parent_run_id)
            if container is None:
                scribe().error(
                    "[MaximSDK][on_llm_start] Couldn't find a container for generation"
                )
                return
            if not container.is_created():
                container.create()
            if container.parent() is None:
                self.containers[str(run_id)] = container
            generation_container = container.add_generation(generation_config)
            # checking if we need to attach evaluator
            if (
                metadata is not None
                and "langgraph_node" in metadata
                and self.eval_config is not None
                and metadata["langgraph_node"] in self.eval_config
            ):
                eval_data = {
                    "container_id": generation_container.id,
                    "generation_container": generation_container,
                    "evaluators": (
                        self.eval_config.get(metadata["langgraph_node"], [])
                        if self.eval_config
                        else []
                    ),
                    "node_name": metadata["langgraph_node"],
                    "input": last_input_message,
                }
                self.to_be_evaluated_container_store.set(
                    str(generation_container.id), eval_data, DEFAULT_TIMEOUT
                )
        except Exception as e:
            import traceback

            scribe().error(
                "[MaximSDK] Failed to process llm-start: %s\n%s",
                e,
                traceback.format_exc(),
            )

    def on_chat_model_start(
        self,
        serialized: Dict[str, Any],
        messages: List[List[BaseMessage]],
        metadata: Optional[Dict[str, Any]] = None,
        **kwargs: Any,
    ) -> Any:
        """Runs when a chat model call starts"""
        try:
            scribe().debug("[MaximSDK: Langchain] on_chat_model_start called")
            self._parse_metadata(metadata)
            run_id = kwargs.get("run_id", None)
            model, model_parameters = parse_langchain_model_parameters(**kwargs)
            provider = parse_langchain_provider(serialized)
            model, provider = parse_langchain_model_and_provider(model, provider)
            maxim_messages = parse_langchain_messages(messages)
            # get user message
            last_input_message = ""
            if maxim_messages is not None:
                for message in maxim_messages:
                    if "role" in message and message["role"] == "user":
                        last_input_message = message["content"]
            generation_id = str(run_id)
            maxim_metadata = self.__get_metadata(metadata)
            generation_name = None
            if kwargs.get("parent_run_id", None) is None:
                generation_name = (
                    maxim_metadata.generation_name if maxim_metadata else None
                )
            # Checking if generation id already present for this run_id
            generation_config = GenerationConfig(
                id=generation_id,
                name=generation_name,
                provider=provider,
                model=model,
                messages=maxim_messages,
                model_parameters=model_parameters,
                tags=maxim_metadata.generation_tags if maxim_metadata else None,
            )
            container = self.__get_container(run_id, kwargs.get("parent_run_id", None))
            if container is None:
                scribe().error(
                    "[MaximSDK][on_chat_model_start] Couldn't find a container for generation]"
                )
                return
            if not container.is_created():
                container.create()
            if container.parent() is None:
                self.containers[str(run_id)] = container
            generation_container = container.add_generation(generation_config)
            # checking if we need to attach evaluator
            if (
                metadata is not None
                and "langgraph_node" in metadata
                and self.eval_config is not None
                and metadata["langgraph_node"] in self.eval_config
            ):
                eval_data = {
                    "container_id": generation_container.id,
                    "generation_container": generation_container,
                    "evaluators": (
                        self.eval_config.get(metadata["langgraph_node"], [])
                        if self.eval_config
                        else []
                    ),
                    "node_name": metadata["langgraph_node"],
                    "input": last_input_message,
                }
                self.to_be_evaluated_container_store.set(
                    str(generation_container.id), eval_data, DEFAULT_TIMEOUT
                )
        except Exception as e:
            import traceback

            scribe().error(
                "[MaximSDK] Failed to process chat-model-start: %s\n%s",
                e,
                traceback.format_exc(),
            )

    def on_llm_new_token(self, token: str, **kwargs: Any) -> Any:
        """Run on new LLM token. Only available when streaming is enabled."""

    def on_llm_end(self, response: LLMResult, **kwargs: Any) -> Any:
        """Run when LLM ends running."""
        try:
            scribe().debug("[MaximSDK][Langchain] on_llm_end called")
            run_id = kwargs.get("run_id", None)
            parent_run_id = kwargs.get("parent_run_id", None)
            result = parse_langchain_llm_result(response)
            container = self.__get_container(run_id, parent_run_id)
            if container is None:
                scribe().error(
                    "[MaximSDK][on_llm_end] Couldn't find a container for generation"
                )
                return
            # here generation_id is the run_id
            self.logger.generation_result(str(run_id), result)
            # We close the container - only if it was created by this run_id
            if container.parent() is None:
                container.end()
                self.containers.pop(str(run_id))

            # check if need to attach evaluator
            if self.to_be_evaluated_container_store.get(str(run_id)):
                obj = self.to_be_evaluated_container_store.get(str(run_id))
                if obj:
                    generation_container: Generation = obj["generation_container"]
                    output = None
                    tool_call_name = None
                    tool_call_args = None
                    # extract output
                    if "choices" in result:
                        if isinstance(result["choices"], list):
                            last_choice = result["choices"][-1]
                            message = last_choice["message"]
                            # check if message content is empty : that means its tool call
                            if len(message["content"]) > 0:
                                output = message["content"]
                            elif len(message["tool_calls"]) > 0:
                                tool_message = message["tool_calls"][-1]
                                if tool_message["type"] == "function":
                                    tool_call_name = tool_message["function"].get(
                                        "name", ""
                                    )
                                    tool_call_args = tool_message["function"].get(
                                        "arguments", ""
                                    )

                    # only keeping non None type variables
                    variable_dict = {
                        k: v
                        for k, v in {
                            "input": obj["input"],
                            "output": output,
                            "tool_call_name": tool_call_name,
                            "tool_call_args": tool_call_args,
                        }.items()
                        if v is not None
                    }

                    generation_container.evaluate().with_evaluators(
                        *obj["evaluators"]
                    ).with_variables(variable_dict)
        except Exception as e:
            import traceback

            scribe().error(
                "[MaximSDK] Failed to process llm-end: %s\n%s",
                e,
                traceback.format_exc(),
            )

    def on_llm_error(
        self, error: Union[Exception, BaseException, KeyboardInterrupt], **kwargs: Any
    ) -> Any:
        """Run when LLM errors."""
        try:
            scribe().debug("[MaximSDK] on_llm_error called")
            run_id = kwargs.get("run_id", None)
            parent_run_id = kwargs.get("parent_run_id", None)
            container = self.__get_container(run_id, parent_run_id)
            if container is None:
                scribe().error(
                    "[MaximSDK][on_llm_error] Couldn't find a container for generation"
                )
                return
            generation_error = parse_langchain_llm_error(error)
            self.logger.generation_error(str(run_id), generation_error)
            # We close the container - only if it was created by this run_id
            if container.parent() is None:
                container.end()
                self.containers.pop(str(run_id))
        except Exception as e:
            import traceback

            scribe().error(
                "[MaximSDK] Failed to process llm-error: %s\n%s",
                e,
                traceback.format_exc(),
            )

    def on_retriever_start(
        self,
        serialized: dict[str, Any],
        query: str,
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        tags: Optional[list[str]] = None,
        metadata: Optional[dict[str, Any]] = None,
        **kwargs: Any,
    ) -> Any:
        """
        Run when Retriever starts running.
        """
        try:
            scribe().debug("[MaximSDK] on_retriever_start called")
            retrieval_id = str(run_id)
            retrieval_config = RetrievalConfig(id=retrieval_id)
            container = self.__get_container(run_id, parent_run_id)
            if container is None:
                scribe().error("[MaximSDK] Couldn't find a container for retrieval")
                return
            if not container.is_created():
                container.create()
            retrieval = container.add_retrieval(retrieval_config)
            retrieval.input(query)
        except Exception as e:
            import traceback

            scribe().error(
                "[MaximSDK] Failed to process retriever-start: %s\n%s",
                e,
                traceback.format_exc(),
            )

    def on_retriever_end(
        self,
        documents: Sequence[Document],
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        **kwargs: Any,
    ) -> Any:
        try:
            scribe().debug("[MaximSDK] on_retriever_end called")
            container = self.__get_container(run_id, parent_run_id)
            if container is None:
                scribe().error("[MaximSDK] Couldn't find a container for retrieval")
                return
            documents_list: List[str] = [doc.page_content for doc in documents]
            self.logger.retrieval_output(str(run_id), documents_list)
            if container.parent() is None:
                container.end()
                self.containers.pop(str(run_id))
        except Exception as e:
            import traceback

            scribe().error(
                "[MaximSDK] Failed to process retriever-end: %s\n%s",
                e,
                traceback.format_exc(),
            )

    # Chain callbacks

    def on_chain_start(
        self, serialized: Dict[str, Any], inputs: Dict[str, Any], **kwargs: Any
    ) -> Any:
        try:
            scribe().debug("[MaximSDK: Langchain] on_chain_start called")
            if "metadata" in kwargs and kwargs.get("metadata", None) is not None:
                self._parse_metadata(kwargs.get("metadata", None))
            run_id = kwargs.get("run_id", None)
            # Creating new span for the chain node
            name = kwargs.get("name", None)
            tags: Dict[str, str] = {}
            if (
                self.metadata is not None
                and self.metadata.get("chain_tags") is not None
            ):
                # its a dict
                for key, value in self.metadata["chain_tags"].items():
                    tags[key] = str(value)
            tags["run_id"] = run_id
            if kwargs.get("parent_run_id", None) is not None:
                tags["parent_run_id"] = str(kwargs.get("parent_run_id", None))
            if kwargs.get("tags", None) is not None:
                for tag in kwargs["tags"]:
                    key, value = tag.split(":", 1)
                    tags[key.strip()] = value.strip()
            if kwargs.get("metadata", None) is not None:
                for key, value in kwargs["metadata"].items():
                    tags[key.strip()] = str(value)

            container = self.__get_container(run_id, kwargs.get("parent_run_id", None))
            if container is None:
                scribe().error("[MaximSDK] Couldn't find a container for chain")
                return

            if not container.is_created():
                container.create()

            if container.parent() is None:
                self.containers[str(run_id)] = container

            span_config = SpanConfig(id=str(run_id), name=name, tags=tags)
            container.add_span(span_config)
            self.containers[str(run_id)] = SpanContainer(
                span_id=str(run_id), logger=self.logger, mark_created=True
            )
        except Exception as e:
            import traceback

            scribe().error(
                "[MaximSDK] Error while processing chain_start %s\n%s",
                e,
                traceback.format_exc(),
            )

    def on_chain_end(self, outputs: Union[str, Dict[str, Any]], **kwargs: Any) -> Any:
        try:
            run_id = kwargs.get("run_id", None)
            parent_run_id = kwargs.get("parent_run_id", None)
            # We hide the hidden chains
            tags = {
                tag.split(":")[0]: tag.split(":")[1]
                for tag in kwargs.get("tags", [])
                if ":" in tag
            }
            container = self.__get_container(run_id, parent_run_id)
            if container is None:
                scribe().error("[MaximSDK] Couldn't find a container for chain")
                return
            container.add_tags(tags)
            container.end()
            self.containers.pop(str(run_id))

        except Exception as e:
            import traceback

            scribe().error(
                "[MaximSDK] Failed to parse chain-end: %s\n%s",
                e,
                traceback.format_exc(),
            )

    def on_chain_error(
        self, error: Union[Exception, BaseException, KeyboardInterrupt], **kwargs: Any
    ) -> Any:
        scribe().info("[MaximSDK] on_chain_error called")
        try:
            run_id = kwargs.get("run_id", None)
            parent_run_id = kwargs.get("parent_run_id", None)
            container = self.__get_container(run_id, parent_run_id)
            if container is None:
                scribe().error("[MaximSDK] Couldn't find a container for chain")
                return
            chain_error = parse_langchain_llm_error(error)
            container.add_error(
                {
                    "id": str(run_id),
                    "message": chain_error.message,
                    "type": chain_error.type,
                    "code": chain_error.code,
                }
            )
        except Exception as e:
            import traceback

            scribe().error(
                "[MaximSDK] Failed to process chain-error: %s\n%s",
                e,
                traceback.format_exc(),
            )

    def on_custom_event(
        self,
        name: str,
        data: Any,
        *,
        run_id: UUID,
        tags: Optional[list[str]] = None,
        metadata: Optional[dict[str, Any]] = None,
        **kwargs: Any,
    ) -> Any:
        try:
            container = self.__get_container(run_id, kwargs.get("parent_run_id", None))
            if container is None:
                scribe().error("[MaximSDK] Couldn't find a container for event")
                return
            final_tags = {
                tag.split(":")[0]: tag.split(":")[1]
                for tag in (tags or [])
                if ":" in tag
            }
            container.add_event(str(run_id), name, final_tags)
        except Exception as e:
            import traceback

            scribe().error(
                "[MaximSDK] Failed to process custom-event: %s\n%s",
                e,
                traceback.format_exc(),
            )

    def on_agent_action(
        self,
        action: AgentAction,
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        **kwargs: Any,
    ) -> Any:
        scribe().warning("[MaximSDK] On agent action is not supported")

    def on_agent_finish(
        self,
        finish: AgentFinish,
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        **kwargs: Any,
    ) -> Any:
        scribe().warning("[MaximSDK] On agent action is not supported")

    # Tool callbacks

    def on_tool_start(
        self, serialized: Dict[str, Any], input_str: str, **kwargs: Any
    ) -> Any:
        try:
            run_id = kwargs.get("run_id", None)
            parent_run_id = kwargs.get("parent_run_id", None)
            name = serialized.get("name", None)
            description = serialized.get("description", None)
            container = self.__get_container(run_id, parent_run_id)
            if container is None:
                scribe().error(
                    "[MaximSDK] Couldn't find a container for this tool call"
                )
                return
            if not container.is_created():
                container.create()
            container.add_tool_call(
                ToolCallConfig(
                    id=str(run_id), name=name, description=description, args=input_str
                )
            )
            if container.parent() is None:
                self.containers[str(run_id)] = container
        except Exception as e:
            import traceback

            scribe().error(
                "[MaximSDK] Failed to parse tool-start: %s\n%s",
                e,
                traceback.format_exc(),
            )

    def on_tool_end(self, output: Any, **kwargs: Any) -> Any:
        try:
            run_id = kwargs.get("run_id", None)
            parent_run_id = kwargs.get("parent_run_id", None)
            container = self.__get_container(run_id, parent_run_id)
            if container is None:
                scribe().error("[MaximSDK] Couldn't find a container for tool call")
                return
            if isinstance(output, ToolMessage):
                result = ""
                if output.content is not None:
                    if isinstance(output.content, str):
                        result = output.content
                    else:
                        result = json.dumps(output.content)
                self.logger.tool_call_result(str(run_id), result)
            elif isinstance(output, dict):
                if output.get("status", None) == "success":
                    self.logger.tool_call_result(
                        str(run_id), output.get("content", None)
                    )
                elif output.get("status", None) == "error":
                    if isinstance(output.get("content", None), str):
                        self.logger.tool_call_error(
                            str(run_id),
                            ToolCallError(message=output.get("content", None)),
                        )
                    else:
                        self.logger.tool_call_error(
                            str(run_id),
                            ToolCallError(message=str(output.get("content", None))),
                        )
                elif output.get("content", None) is not None:
                    self.logger.tool_call_result(
                        str(run_id), output.get("content", None)
                    )

            if container.parent() is None:
                container.end()
                self.containers.pop(str(run_id))
        except Exception as e:
            import traceback

            scribe().error(
                "[MaximSDK] Failed to parse tool-end: %s\n%s",
                e,
                traceback.format_exc(),
            )

    def on_tool_error(
        self, error: Union[Exception, BaseException, KeyboardInterrupt], **kwargs: Any
    ) -> Any:
        try:
            run_id = kwargs.get("run_id", None)
            container = self.__get_container(run_id, kwargs.get("parent_run_id", None))
            if container is None:
                scribe().error("[MaximSDK] Couldn't find a container for tool_call")
                return
            if error is not None and hasattr(error, "__str__"):
                self.logger.tool_call_error(
                    str(run_id), ToolCallError(message=str(error))
                )
            if container.parent() is None:
                container.end()
                self.containers.pop(str(run_id))
        except Exception as e:
            import traceback

            scribe().error(
                "[MaximSDK] Failed to parse tool-end: %s\n%s",
                e,
                traceback.format_exc(),
            )

    def on_text(self, text: str, **kwargs: Any) -> Any:
        scribe().warning("[MaximSDK] Text models are not supported by Maxim callback")



import copy
import functools
import inspect
import json
import time
import traceback
from io import BytesIO
from typing import Union

from google.genai.types import (
    LiveConnectConfig,
    LiveServerContent,
    LiveServerToolCall,
    UsageMetadata,
    Content,
)
from livekit.agents.llm import InputTranscriptionCompleted
from livekit.plugins.google.beta.realtime.realtime_api import RealtimeSession
from livekit.rtc import AudioFrame

from ....scribe import scribe
from ...components import (
    AudioContent,
    FileDataAttachment,
    GenerationResult,
    GenerationResultChoice,
    ImageContent,
    TextContent,
    TokenDetails,
)
from ...utils import pcm16_to_wav_bytes
from ..store import (
    SessionStoreEntry,
    get_maxim_logger,
    get_session_store,
)
from ..utils import get_thread_pool_executor


def handle_build_connect_config_result(
    self: RealtimeSession, result: LiveConnectConfig
):
    # here we will start the new generation
    # this is same as session started
    session_info = get_session_store().get_session_by_rt_session_id(id(self))
    if session_info is None:
        scribe().warning(
            "[MaximSDK] session info is none at realtime session emit. If you are seeing this frequently, please report issue at https://github.com/maximhq/maxim-py/issues"
        )
        return
    session_info.provider = "google-realtime"
    llm_config = {}
    if result.generation_config is not None:
        llm_config = result.generation_config.model_dump()
    if result.speech_config is not None:
        for key, value in result.speech_config.model_dump().items():
            if value is not None:
                llm_config[key] = value
    # removing speech_config from the llm_config
    llm_config.pop("speech_config", None)
    llm_config["model"] = self._opts.model
    # checking if there are any tools, add adding them in the llm_config
    if result.tools is not None and len(result.tools) > 0:
        llm_config["tools"] = []
        # Third party tools
        for tool in result.tools:
            if tool.function_declarations is None:
                continue
            for func_declaration in tool.function_declarations:
                if func_declaration is None:
                    continue
                llm_config["tools"].append(
                    {
                        "name": func_declaration.name,
                        "description": func_declaration.description,
                        "parameters": func_declaration.parameters,
                    }
                )
    # saving back llm_config
    session_info.llm_config = llm_config
    # saving back the session
    get_session_store().set_session(session_info)
    trace_id = session_info.mx_current_trace_id
    if trace_id is None:
        return
    trace = get_maxim_logger().trace({"id": trace_id})
    turn = session_info.current_turn
    if turn is None:
        return
    system_prompt = ""
    if (
        result.system_instruction is not None
        and isinstance(result.system_instruction, Content)
        and result.system_instruction.parts is not None
    ):
        if result.system_instruction.parts is not None:
            # We have to iterate through the system instructions
            for part in result.system_instruction.parts:
                if part.text is not None:
                    system_prompt += part.text
    trace.generation(
        {
            "id": turn.turn_id,
            "model": self._opts.model,
            "name": "LLM call",
            "provider": "google",
            "model_parameters": llm_config,
            "messages": [{"role": "system", "content": system_prompt}],
        }
    )
    session_info.user_speaking = False
    get_session_store().set_session(session_info)


def handle_server_content(self: RealtimeSession, content: LiveServerContent):
    session_info = get_session_store().get_session_by_rt_session_id(id(self))
    if session_info is None:
        scribe().warning(
            "[MaximSDK] session info is none at realtime session emit. If you are seeing this frequently, please report issue at https://github.com/maximhq/maxim-py/issues"
        )
        return
    turn = session_info.current_turn
    if turn is None:
        return
    if content.output_transcription is not None and session_info.user_speaking:
        session_info.user_speaking = False
        get_session_store().set_session(session_info)
    # adding transcription
    if (
        content.output_transcription is not None
        and content.output_transcription.text is not None
    ):
        turn.turn_output_transcription += content.output_transcription.text
    # getting audio bytes from the payload
    if (
        content.model_turn is not None
        and content.model_turn.parts is not None
        and len(content.model_turn.parts) > 0
    ):
        for part in content.model_turn.parts:
            if part.inline_data is None or part.inline_data.data is None:
                continue
            turn.turn_output_audio_buffer.write(part.inline_data.data)
            if (
                session_info.conversation_buffer.tell() + len(part.inline_data.data)
                > 10 * 1024 * 1024
            ):
                session_id = session_info.mx_session_id
                if session_id is None:
                    return
                index = session_info.conversation_buffer_index
                get_maxim_logger().session_add_attachment(
                    session_id,
                    FileDataAttachment(
                        data=pcm16_to_wav_bytes(
                            session_info.conversation_buffer.getvalue()
                        ),
                        tags={"attach-to": "input"},
                        name=f"Conversation part {index}",
                        timestamp=int(time.time()),
                    ),
                )
                session_info.conversation_buffer = BytesIO()
                session_info.conversation_buffer_index = index + 1
            session_info.conversation_buffer.write(part.inline_data.data)
    session_info.current_turn = turn
    get_session_store().set_session(session_info)


def handle_google_input_transcription_completed(
    session_info: SessionStoreEntry, input_transcription: InputTranscriptionCompleted
):
    turn = session_info.current_turn
    if turn is None:
        return
    turn.turn_input_transcription = input_transcription.transcript
    get_session_store().set_session(session_info)


def handle_usage_metadata(self: RealtimeSession, usage: UsageMetadata):
    session_info = get_session_store().get_session_by_rt_session_id(id(self))
    if session_info is None:
        scribe().warning(
            "[MaximSDK] session info is none at realtime session emit. If you are seeing this frequently, please report issue at https://github.com/maximhq/maxim-py/issues"
        )
        return
    # Here if the interrupted turn is present, we need to use that instead of the current turn
    turn = session_info.current_turn
    if turn is None:
        return
    # Writing final Generation result
    if (
        turn.turn_input_audio_buffer is not None
        and turn.turn_input_audio_buffer.tell() > 0
    ):
        buffer = turn.turn_input_audio_buffer.getvalue()
        # Remove last 5 seconds of audio (assuming 16kHz, 16-bit mono PCM)
        sample_rate = 16000
        sample_width = 2  # bytes per sample for 16-bit PCM
        seconds_to_remove = 5
        bytes_to_remove = sample_rate * sample_width * seconds_to_remove
        if len(buffer) > bytes_to_remove:
            buffer = buffer[:-bytes_to_remove]
        else:
            buffer = b""
        get_maxim_logger().generation_add_attachment(
            turn.turn_id,
            FileDataAttachment(
                data=pcm16_to_wav_bytes(buffer),
                tags={"attach-to": "input"},
                name="User Input",
                timestamp=int(time.time()),
            ),
        )
    if (
        turn.turn_output_audio_buffer is not None
        and turn.turn_output_audio_buffer.tell() > 0
    ):
        get_maxim_logger().generation_add_attachment(
            turn.turn_id,
            FileDataAttachment(
                data=pcm16_to_wav_bytes(turn.turn_output_audio_buffer.getvalue()),
                tags={"attach-to": "output"},
                name="Assistant Response",
                timestamp=int(time.time()),
            ),
        )
    contents: list[Union[TextContent, ImageContent, AudioContent]] = []
    contents.append({"type": "audio", "transcript": turn.turn_output_transcription})
    choices: list[GenerationResultChoice] = []
    choices.append(
        {
            "index": 0,
            "logprobs": None,
            "finish_reason": "stop",
            "message": {"role": "assistant", "content": contents, "tool_calls": []},
        }
    )
    get_maxim_logger().generation_set_provider(turn.turn_id, "google")
    # Parsing token details
    input_token_details: dict[str, int] = {}
    output_token_details: dict[str, int] = {}
    cached_token_details: dict[str, int] = {}
    if usage.prompt_tokens_details is not None:
        for detail in usage.prompt_tokens_details:
            if detail.modality == "TEXT" and detail.token_count is not None:
                input_token_details["text_tokens"] = detail.token_count
            elif detail.modality == "AUDIO" and detail.token_count is not None:
                input_token_details["audio_tokens"] = detail.token_count
    if usage.response_tokens_details is not None:
        for detail in usage.response_tokens_details:
            if detail.modality == "TEXT" and detail.token_count is not None:
                output_token_details["text_tokens"] = detail.token_count
            elif detail.modality == "AUDIO" and detail.token_count is not None:
                output_token_details["audio_tokens"] = detail.token_count
    if usage.cache_tokens_details is not None:
        for detail in usage.cache_tokens_details:
            if detail.modality == "TEXT" and detail.token_count is not None:
                cached_token_details["text_tokens"] = detail.token_count
            elif detail.modality == "AUDIO" and detail.token_count is not None:
                cached_token_details["audio_tokens"] = detail.token_count
    if session_info.rt_session_id is None or session_info.mx_current_trace_id is None:
        return
    trace = get_session_store().get_current_trace_from_rt_session_id(
        session_info.rt_session_id
    )
    get_maxim_logger().trace_add_generation(
        session_info.mx_current_trace_id,
        {
            "id": turn.turn_id,
            "model": self._opts.model,
            "name": "LLM call",
            "provider": "google",
            "messages": [{"role": "user", "content": turn.turn_input_transcription}],
            "model_parameters": session_info.llm_config,  # type: ignore
        },
    )
    result: GenerationResult = {
        "id": turn.turn_id,
        "object": "",
        "created": int(time.time()),
        "model": self._opts.model,
        "usage": {
            "completion_tokens": (
                usage.response_token_count
                if usage.response_token_count is not None
                else 0
            ),
            "prompt_tokens": (
                usage.prompt_token_count if usage.prompt_token_count is not None else 0
            ),
            "total_tokens": (
                usage.total_token_count if usage.total_token_count is not None else 0
            ),
            "input_token_details": TokenDetails(**input_token_details),
            "output_token_details": TokenDetails(**output_token_details),
            "cached_token_details": TokenDetails(**cached_token_details),
        },
        "choices": choices,
    }
    # Setting up the generation
    get_maxim_logger().generation_result(turn.turn_id, result)
    # Setting the output to the trace
    if session_info.rt_session_id is not None:
        trace = get_session_store().get_current_trace_from_rt_session_id(
            session_info.rt_session_id
        )
        if (
            trace is not None
            and len(choices) > 0
            and choices[0]["message"]["content"] is not None
            and isinstance(choices[0]["message"]["content"], list)
            and len(choices[0]["message"]["content"]) > 0
            and choices[0]["message"]["content"][0] is not None
            and "transcript" in choices[0]["message"]["content"][0]
        ):
            trace.set_output(choices[0]["message"]["content"][0]["transcript"])


def handle_push_audio(self: RealtimeSession, data_buffer: bytes):
    session_info = get_session_store().get_session_by_rt_session_id(id(self))
    if session_info is None:
        scribe().warning(
            "[MaximSDK] session info is none at realtime session emit. If you are seeing this frequently, please report issue at https://github.com/maximhq/maxim-py/issues"
        )
        return
    turn = session_info.current_turn
    if turn is None:
        return
    # This will help us skip silence before the turn starts
    if (data_buffer is None or len(data_buffer) == 0) and (
        turn.turn_input_audio_buffer is None or turn.turn_input_audio_buffer.tell() == 0
    ):
        return
    # Ensure we have a valid audio buffer
    if turn.turn_input_audio_buffer is None:
        turn.turn_input_audio_buffer = BytesIO()
    turn.turn_input_audio_buffer.write(data_buffer)
    session_info.conversation_buffer.write(data_buffer)
    session_info.current_turn = turn
    get_session_store().set_session(session_info)


def handle_tool_calls(self, tool_calls: LiveServerToolCall):
    """
    This function is called when the agent makes a tool call.
    """
    trace = get_session_store().get_current_trace_from_rt_session_id(id(self))
    if trace is None:
        return
    if tool_calls is None:
        return
    if tool_calls.function_calls is None:
        return
    for function_call in tool_calls.function_calls:
        if function_call is None:
            continue
        if function_call.id is None:
            continue
        raw_args = function_call.args
        if isinstance(raw_args, (dict, list)):
            raw_args = json.dumps(raw_args)
        trace.tool_call(
            {
                "id": function_call.id,
                "name": (
                    function_call.name
                    if function_call.name is not None
                    else "tool_call"
                ),
                "args": raw_args if isinstance(raw_args, str) else json.dumps(raw_args),
            }
        )


ignored_hooks = ["_resample_audio", "_send_client_event", "_start_new_generation"]


def pre_hook(self: RealtimeSession, hook_name, args, kwargs):
    try:
        if hook_name in ignored_hooks:
            return
        elif hook_name == "_handle_server_content":
            # can we do deep copy of args[0]
            # and then pass it to the thread pool executor
            get_thread_pool_executor().submit(
                handle_server_content, self, copy.deepcopy(args[0])
            )
        elif hook_name == "_handle_tool_calls":
            get_thread_pool_executor().submit(handle_tool_calls, self, args[0])
        elif hook_name == "_handle_usage_metadata":
            get_thread_pool_executor().submit(
                handle_usage_metadata, self, copy.deepcopy(args[0])
            )
        elif hook_name == "push_audio":
            if not args or len(args) == 0:
                return
            audio_frame: AudioFrame = args[0]
            data_buffer = memoryview(audio_frame._data).cast("B")
            get_thread_pool_executor().submit(handle_push_audio, self, data_buffer)
        elif hook_name == "_mark_current_generation_done":
            scribe().debug(
                f"[Internal][Gemini:{self.__class__.__name__}] _mark_current_generation_done called; args={args}, kwargs={kwargs}"
            )
        else:
            scribe().debug(
                f"[Internal][Gemini:{self.__class__.__name__}] {hook_name} called; args={args}, kwargs={kwargs}"
            )
    except Exception as e:
        scribe().warning(
            f"[Internal][Gemini:{self.__class__.__name__}] {hook_name} failed; error={e!s}\n{traceback.format_exc()}"
        )


def post_hook(self: RealtimeSession, result, hook_name, args, kwargs):
    try:
        if hook_name in ignored_hooks:
            return
        if hook_name in (
            "_resample_audio",
            "push_audio",
            "_send_client_event",
            "_send_task",
            "_start_new_generation",
            "_handle_server_content",
            "_handle_usage_metadata",
        ):
            return
        if hook_name == "_build_connect_config":
            handle_build_connect_config_result(self, result)
        else:
            scribe().debug(
                f"[Internal][Gemini:{self.__class__.__name__}] {hook_name} completed; result={result}"
            )
    except Exception as e:
        scribe().warning(
            f"[Internal][Gemini:{self.__class__.__name__}] {hook_name} failed; error={e!s}\n{traceback.format_exc()}"
        )


def instrument_gemini_session(orig, name):
    if inspect.iscoroutinefunction(orig):

        async def async_wrapper(self, *args, **kwargs):
            pre_hook(self, name, args, kwargs)
            result = None
            try:
                result = await orig(self, *args, **kwargs)
                return result
            finally:
                post_hook(self, result, name, args, kwargs)

        wrapper = async_wrapper
    else:

        def sync_wrapper(self, *args, **kwargs):
            pre_hook(self, name, args, kwargs)
            result = None
            try:
                result = orig(self, *args, **kwargs)
                return result
            finally:
                post_hook(self, result, name, args, kwargs)

        wrapper = sync_wrapper
    return functools.wraps(orig)(wrapper)



import json
import logging
import re
import time
from typing import Any, Dict, List, Tuple, Union, Optional
from uuid import uuid4

from langchain_core.messages import AIMessage, BaseMessage, ToolCall
from langchain_core.outputs import ChatResult, LLMResult
from langchain_core.outputs.chat_generation import ChatGeneration, ChatGenerationChunk
from langchain_core.outputs.generation import Generation, GenerationChunk

from ...scribe import scribe
from ..components.types import GenerationError

logger = logging.getLogger("MaximSDK")


def parse_langchain_model_and_provider(model: str, provider: str) -> Tuple[str, str]:
    return model, provider


def parse_langchain_provider(serialized: Dict[str, Any]):
    """Parses langchain provider from serialized data
    Args:
        serialized: Dict[str, Any]: Serialized data to parse provider from
    Returns:
        str: Parsed provider
    """
    provider = serialized.get("name", "").lower()
    if provider.startswith("chat"):
        return provider.replace("chat", "")
    mapping = {
        "azure": "azure",
        "anthropic": "anthropic",
        "huggingface": "huggingface",
        "bedrock": "bedrock",
        "aws": "bedrock",
        "openai": "openai",
        "groq": "groq",
        "ollama": "ollama",
        "gemini": "google",
        "vertexai": "google",
        "deepseek": "deepseek",
        "qwen": "qwen",
    }
    for key, target in mapping.items():
        if key in provider:
            return target
    return "unknown"


def parse_langchain_llm_error(
    error: Union[Exception, BaseException, KeyboardInterrupt],
) -> GenerationError:
    """Parses langchain LLM error into a format that is accepted by Maxim logger
    Args:
        error: Union[Exception, KeyboardInterrupt]: Error to be parsed
    Returns:
        GenerationError: Parsed LLM error
    """
    if isinstance(error, KeyboardInterrupt):
        return GenerationError(message="Generation was interrupted by the user")
    if isinstance(error, Exception):
        return GenerationError(message=str(error))
    else:
        message = error.__dict__.get("message", "")
        type = error.__dict__.get("type", None)
        code = error.__dict__.get("code", None)
        return GenerationError(message=message, type=type, code=code)


def parse_langchain_model_parameters(**kwargs: Any) -> Tuple[str, Dict[str, Any]]:
    """Parses langchain kwargs into model and model parameters. You can use this function with any langchain _start callback function
    Args:
        kwargs: Dict[str, Any]: Kwargs to be parsed
    Returns:
        Tuple[str, Dict[str, Any]]: Model and model parameters
    Raises:
        Exception: If model_name is not found in kwargs
    """
    model_parameters = kwargs.get("invocation_params", {})
    # Checking if model_name present
    model = "unknown"
    if "model_name" in model_parameters:
        model = model_parameters["model_name"]
        del model_parameters["model_name"]
    # If not then checking if invocation_params exist
    elif "model" in model_parameters:
        model = model_parameters["model"]
        del model_parameters["model"]
    elif "model_id" in model_parameters:
        model = model_parameters["model_id"]
        del model_parameters["model_id"]
    return model, model_parameters


def parse_base_message_to_maxim_generation(message: BaseMessage):
    """Parses langchain BaseMessage into a format that is accepted by Maxim logger
    Args:
        message: BaseMessage
    Returns:
        Dict[str, Any]: Parsed message
    """
    choice = parse_langchain_message(message)
    usage = (
        message.__dict__["usage_metadata"] if message.__dict__["usage_metadata"] else {}
    )
    return {
        "id": str(uuid4()),
        "created": int(time.time()),
        "choices": [choice],
        "usage": {
            "prompt_tokens": usage.get("input_tokens", 0),
            "completion_tokens": usage.get("output_tokens", 0),
            "total_tokens": usage.get("total_tokens", 0),
        },
    }


def parse_langchain_message(message: BaseMessage):
    """Parses langchain BaseMessage into a choice of openai message
    Args:
        message: BaseMessage
    Returns:
        Dict[str, Any]: Parsed message
    """
    message_type = message.__dict__["type"]
    if message_type == "ai":
        ai_message = AIMessage(
            content=message.content, additional_kwargs=message.additional_kwargs
        )
        tool_calls = (
            ai_message.tool_calls or message.lc_attributes.get("tool_calls") or []
        )
        return {
            "index": 0,
            "message": {
                "role": "assistant",
                "content": (
                    message.content
                    if isinstance(message.content, str)
                    else json.dumps(message.content)
                ),
                "tool_calls": parse_langchain_tool_call(tool_calls),
            },
            "finish_reason": (
                message.response_metadata.get("finish_reason")
                or message.response_metadata.get("stop_reason")
                or "stop"
            ),
            "logprobs": (message.response_metadata.get("logprobs") or None),
        }
    else:
        return {
            "index": 0,
            "message": {
                "role": "assistant",
                "content": message.content,
            },
            "finish_reason": (
                message.response_metadata["finish_reason"]
                if message.response_metadata["finish_reason"]
                else None
            ),
            "logprobs": (
                message.response_metadata["logprobs"]
                if message.response_metadata["logprobs"]
                else None
            ),
        }


def parse_langchain_tool_call(tool_calls: List[ToolCall]):
    final_tool_calls = []
    for tool_call in tool_calls:
        try:
            final_tool_calls.append(
                {
                    "type": "function",
                    "id": tool_call.get("id", str(uuid4())),
                    "function": {
                        "name": tool_call.get("name", "unknown"),
                        "arguments": json.dumps(tool_call.get("args", {})),
                    },
                }
            )
        except AttributeError as e:
            scribe().debug(f"Error parsing tool call: {str(e)}")
            # trying the dict way
            dict_tool_call = tool_call.__dict__
            final_tool_calls.append(
                {
                    "type": "function",
                    "id": dict_tool_call.get("id", str(uuid4())),
                    "function": {
                        "name": dict_tool_call.get("name", "unknown"),
                        "arguments": json.dumps(dict_tool_call.get("args", {})),
                    },
                }
            )
        except Exception as e:
            scribe().error(f"Error parsing tool call: {str(e)}")
    return final_tool_calls


def parse_langchain_chat_generation_chunk(generation: ChatGeneration):
    choices = []
    content = generation.text
    finish_reason = (
        generation.message.response_metadata.get("stop_reason")
        if generation.message.response_metadata
        else "stop"
    )
    if finish_reason is None:
        finish_reason = (
            generation.generation_info.get("finish_reason")
            if generation.generation_info
            else "stop"
        )
    choices.append(
        {
            "index": 0,
            "message": {"role": "assistant", "content": content, "tool_calls": []},
            "finish_reason": finish_reason,
            "logprobs": (
                generation.generation_info.get("logprobs")
                if generation.generation_info
                else None
            ),
        }
    )
    return choices


def get_action_from_kwargs(kwargs: Dict[str, Any]) -> Optional[Dict]:
    """
    Extract action from tool outputs in kwargs.

    Args:
        kwargs: Dictionary containing tool outputs

    Returns:
        The action dict if found, None otherwise
    """
    try:
        tool_outputs = kwargs.get("tool_outputs", [])
        if tool_outputs and isinstance(tool_outputs, list) and len(tool_outputs) > 0:
            first_output = tool_outputs[0]
            if isinstance(first_output, dict) and "action" in first_output:
                return first_output["action"]
        return None
    except (KeyError, IndexError, TypeError) as e:
        return None


def parse_langchain_chat_generation(generation: ChatGeneration):
    choices = []
    message = generation.message
    if message.type == "ai":
        ai_message = AIMessage(
            content=message.content, additional_kwargs=message.additional_kwargs
        )
        finish_reason = (
            generation.message.response_metadata.get("stop_reason")
            if generation.message.response_metadata
            else "stop"
        )
        if finish_reason is None:
            finish_reason = (
                generation.generation_info.get("finish_reason")
                if generation.generation_info
                else "stop"
            )
        tool_calls = (
            ai_message.tool_calls or message.lc_attributes.get("tool_calls") or []
        )
        content = ""
        actions = get_action_from_kwargs(ai_message.additional_kwargs)
        actions_dict = {"Action": actions}
        if actions:
            content += json.dumps(actions_dict)
        if isinstance(ai_message.content, str):
            content = ai_message.content
        elif isinstance(ai_message.content, list):
            for item in ai_message.content:
                if isinstance(item, str):
                    content += item
                elif isinstance(item, dict):
                    if "type" in item and item["type"] == "text":
                        content += item["text"]
                    elif "type" in item and item["type"] == "image_url":
                        content += item["image_url"]
                    else:
                        content += json.dumps(item)
        else:
            content = json.dumps(ai_message.content)
        choices.append(
            {
                "index": 0,
                "message": {
                    "role": "assistant",
                    "content": content,
                    "tool_calls": parse_langchain_tool_call(tool_calls),
                },
                "finish_reason": finish_reason,
                "logprobs": (
                    generation.generation_info.get("logprobs")
                    if generation.generation_info
                    else None
                ),
            }
        )
    return choices


def parse_langchain_generation_chunk(generation: GenerationChunk):
    return [
        {
            "index": 0,
            "text": generation.text,
            "logprobs": (
                generation.generation_info.get("logprobs")
                if generation.generation_info
                else None
            ),
            "finish_reason": (
                generation.generation_info.get("finish_reason")
                if generation.generation_info
                else "stop"
            ),
        }
    ]


def parse_langchain_text_generation(generation: Generation):
    choices = []
    messages = parse_langchain_messages([generation.text], "system")
    if len(messages) > 0:
        for i, message in enumerate(messages):
            choices.append(
                {
                    "index": i,
                    "text": message["content"],
                    "logprobs": (
                        generation.generation_info.get("logprobs")
                        if generation.generation_info
                        else None
                    ),
                    "finish_reason": (
                        generation.generation_info.get("finish_reason")
                        if generation.generation_info
                        else None
                    ),
                }
            )
    return choices


def parse_langchain_generation(generation: Generation):
    """Parses langchain generation into a format that is accepted by Maxim logger
    Args:
        generation: Generation: Generation to be parsed
    Returns:
        Dict[str, Any]: Parsed generation
    """
    # Sequence of checks matter here as ChatGenerationChunk is a subclass of ChatGeneration
    if isinstance(generation, ChatGenerationChunk):
        scribe().debug("[MaximSDK][Langchain] Parsing ChatGenerationChunk")
        return parse_langchain_chat_generation_chunk(generation)
    elif isinstance(generation, ChatGeneration):
        scribe().debug("[MaximSDK][Langchain] Parsing ChatGeneration")
        return parse_langchain_chat_generation(generation)
    elif isinstance(generation, GenerationChunk):
        scribe().debug("[MaximSDK][Langchain] Parsing GenerationChunk")
        return parse_langchain_generation_chunk(generation)
    elif isinstance(generation, Generation):
        scribe().debug("[MaximSDK][Langchain] Parsing Generation")
        return parse_langchain_text_generation(generation)


def parse_token_usage_for_result(result: LLMResult):
    """
    Parses token usage for a given LLM result
    """
    usage = result.llm_output.get("token_usage") if result.llm_output else None
    if usage is not None:
        return usage
    llm_usage = result.llm_output.get("usage") if result.llm_output else None
    if llm_usage:
        if llm_usage.get("input_tokens") is not None:
            usage = {
                "prompt_tokens": llm_usage.get("input_tokens", 0),
                "completion_tokens": llm_usage.get("output_tokens", 0),
                "total_tokens": llm_usage.get("input_tokens", 0)
                + llm_usage.get("output_tokens", 0),
            }
        elif llm_usage.get("prompt_tokens") is not None:
            usage = {
                "prompt_tokens": llm_usage.get("prompt_tokens", 0),
                "completion_tokens": llm_usage.get("completion_tokens", 0),
                "total_tokens": llm_usage.get("prompt_tokens", 0)
                + llm_usage.get("completion_tokens", 0),
            }
    if usage is not None:
        return usage
    # Here we might have to go down to each generation and sum up all usages
    prompt_tokens = 0
    output_tokens = 0
    total_tokens = 0
    generations = result.generations
    if generations is not None:
        for _, generation in enumerate(generations):
            if generation is None:
                continue
            for _, gen in enumerate(generation):
                if gen is None or isinstance(gen, str):
                    continue
                usage_data = None
                if "message" in gen.__dict__:
                    message_obj = gen.__dict__.get("message")
                    if message_obj and hasattr(message_obj, "__dict__"):
                        usage_data = message_obj.__dict__.get("usage_metadata")
                elif (
                    "generation_info" in gen.__dict__
                    and gen.__dict__.get("generation_info", None) is not None
                ):
                    usage_data = gen.__dict__.get("generation_info", {}).get(
                        "usage_metadata"
                    )
                if usage_data is not None:
                    if usage_data.get("input_tokens") is not None:
                        prompt_tokens += usage_data.get("input_tokens", 0)
                        output_tokens += usage_data.get("output_tokens", 0)
                        total_tokens += usage_data.get(
                            "input_tokens", 0
                        ) + usage_data.get("output_tokens", 0)
                        continue
                    elif usage_data.get("prompt_tokens") is not None:
                        prompt_tokens += usage_data.get("prompt_tokens", 0)
                        output_tokens += usage_data.get("completion_tokens", 0)
                        total_tokens += usage_data.get(
                            "prompt_tokens", 0
                        ) + usage_data.get("completion_tokens", 0)
                        continue
                    elif usage_data.get("prompt_token_count") is not None:
                        prompt_tokens += usage_data.get("prompt_token_count", 0)
                        output_tokens += usage_data.get("candidates_token_count", 0)
                        total_tokens += usage_data.get(
                            "prompt_token_count", 0
                        ) + usage_data.get("completion_token_count", 0)
                        continue
                message_obj = gen.__dict__.get("message")
                resp_metadata = None
                if message_obj and hasattr(message_obj, "__dict__"):
                    resp_metadata = message_obj.__dict__.get("response_metadata")
                if resp_metadata is not None:
                    usage_data = resp_metadata.get("usage") or None
                    if usage_data is not None:
                        if usage_data.get("input_tokens") is not None:
                            prompt_tokens += usage_data.get("input_tokens", 0)
                            output_tokens += usage_data.get("output_tokens", 0)
                            total_tokens += usage_data.get(
                                "input_tokens", 0
                            ) + usage_data.get("output_tokens", 0)
                            continue
                        elif usage_data.get("prompt_tokens") is not None:
                            prompt_tokens += usage_data.get("prompt_tokens", 0)
                            output_tokens += usage_data.get("completion_tokens", 0)
                            total_tokens += usage_data.get(
                                "prompt_tokens", 0
                            ) + usage_data.get("completion_tokens", 0)
                            continue
                    # The last case we check for is bedrock
                    usage_data = (
                        resp_metadata.get("amazon-bedrock-invocationMetrics")
                        if resp_metadata
                        else None
                    )
                    if usage_data is not None:
                        prompt_tokens += usage_data.get("inputTokenCount", 0)
                        output_tokens += usage_data.get("outputTokenCount", 0)
                        total_tokens += usage_data.get(
                            "inputTokenCount", 0
                        ) + usage_data.get("outputTokenCount", 0)
    return {
        "prompt_tokens": prompt_tokens,
        "completion_tokens": output_tokens,
        "total_tokens": total_tokens,
    }


def parse_langchain_chat_result(result: ChatResult) -> Dict[str, Any]:
    """Parses langchain Chat result into a format that is accepted by Maxim logger
    Args:
        result: ChatResult: Chat result to be parsed
    Returns:
        Dict[str, Any]: Parsed Chat result
    Raises:
        Exception: If error parsing Chat result
    """
    try:
        generations = result.generations
        choices = []
        if generations is not None:
            for _, generation in enumerate(generations):
                for _, gen in enumerate(generation):
                    parsed_generations = parse_langchain_generation(gen)
                    if isinstance(parsed_generations, list):
                        choices.extend(parsed_generations)
                    else:
                        choices.append(parsed_generations)
        usage = parse_token_usage_for_result(result)
        # Adding index to each choice
        for i, choice in enumerate(choices):
            choices[i] = {**choice, "index": i}
        return {
            "id": str(uuid4()),
            "created": int(time.time()),
            "choices": choices,
            "usage": usage,
        }
    except Exception as e:
        logger.error(f"Error parsing LLM result: {str(e)}")
        raise Exception(f"Error parsing LLM result: {str(e)}") from e


def parse_langchain_llm_result(result: LLMResult) -> Dict[str, Any]:
    """Parses langchain LLM result into a format that is accepted by Maxim logger
    Args:
        result: LLMResult: LLM result to be parsed
    Returns:
        Dict[str, Any]: Parsed LLM result
    Raises:
        Exception: If error parsing LLM result
    """
    try:
        generations = result.generations
        choices = []
        if generations is not None:
            for _, generation in enumerate(generations):
                for _, gen in enumerate(generation):
                    parsed_generations = parse_langchain_generation(gen)
                    if isinstance(parsed_generations, list):
                        choices.extend(parsed_generations)
                    else:
                        choices.append(parsed_generations)
        usage = parse_token_usage_for_result(result)
        # Adding index to each choice
        for i, choice in enumerate(choices):
            choices[i] = {**choice, "index": i}
        return {
            "id": str(uuid4()),
            "created": int(time.time()),
            "choices": choices,
            "usage": usage,
        }
    except Exception as e:
        logger.error(f"Error parsing LLM result: {str(e)}")
        raise Exception(f"Error parsing LLM result: {str(e)}")


def parse_langchain_messages(
    input: Union[List[str], List[List[Any]]], default_role="user"
):
    """Parses langchain messages into messages that are accepted by Maxim logger
    Args:
        input: List[str] or List[List[Any]]: List of messages to be parsed
        default_role: str: Default role to assign to messages without a role
    Returns:
        List[Dict[str, str]]: List of messages with role and content
    Raises:
        Exception: If input is not List[str] or List[List[Any]]
        Exception: If message type is not str or list
        Exception: If message type is not recognized
    """
    try:
        delimiter_to_role = {
            "System": "system",
            "Human": "user",
            "User": "user",
            "Assistant": "assistant",
            "Model": "model",
            "Tool": "tool",
        }
        messages = []
        # checking if input is List[str] or List[List]
        if isinstance(input[0], list):
            for message_list in input or []:
                for message in message_list:
                    if isinstance(message, str):
                        messages.append({"role": default_role, "content": message})
                        continue
                    message_type = type(message).__name__
                    if message_type.endswith("SystemMessage"):
                        messages.append(
                            {"role": "system", "content": message.content or ""}
                        )
                    elif message_type.endswith("HumanMessage"):
                        messages.append(
                            {"role": "user", "content": message.content or ""}
                        )
                    elif message_type.endswith("AIMessage"):
                        messages.append(
                            {"role": "assistant", "content": message.content or ""}
                        )
                    elif message_type.endswith("ToolMessage"):
                        messages.append(
                            {
                                "role": "tool",
                                "content": message.content or "",
                                "tool_call_id": message.tool_call_id,
                            }
                        )
                    elif message_type.endswith("ChatMessage"):
                        messages.append(
                            {
                                "role": message.role,
                                "content": message.content or "",
                            }
                        )
                    else:
                        logger.error(f"Invalid message type: {type(message)}")
                        raise Exception(f"Invalid message type: {type(message)}")
        else:
            for message in input or []:
                if not isinstance(message, str):
                    logger.error(f"Invalid message type: {type(message)}")
                    raise Exception(f"Invalid message type: {type(message)}")
                # get type of message
                # Define the delimiter pattern
                pattern = r"(System:|Human:|User:|Assistant:|Model:|Tool:)"
                # Split the text using the pattern
                splits = re.split(pattern, message)
                # Remove any leading/trailing whitespace and empty strings
                splits = [s.strip() for s in splits if s.strip()]
                # Pair up the delimiters with their content
                for i in range(0, len(splits), 2):
                    if i + 1 < len(splits):
                        # Remove ":" from delimiter and trim both delimiter and content
                        delimiter = splits[i].rstrip(":").strip()
                        content = splits[i + 1].strip()
                        messages.append(
                            {
                                "role": delimiter_to_role.get(delimiter, "user"),
                                "content": content,
                            }
                        )
                    else:
                        if splits[i].find(":") == -1:
                            messages.append(
                                {
                                    "role": delimiter_to_role.get(default_role, "user"),
                                    "content": splits[i],
                                }
                            )
                        else:
                            # Handle case where there's a delimiter without content
                            delimiter = splits[i].rstrip(":").strip()
                            messages.append(
                                {
                                    "role": delimiter_to_role.get(delimiter, "user"),
                                    "content": "",
                                }
                            )
        return messages
    except Exception as e:
        logger.error(f"Error parsing messages: {e}")
        raise Exception(f"Error parsing messages: {e}")



"""Utility functions for Together AI API integration with Maxim.

This module provides utility functions for parsing and processing Together AI API
requests and responses to integrate with Maxim's logging and monitoring system.
It handles message format conversion, parameter extraction, response standardization,
and image attachment processing.
"""

import time
import uuid
from typing import Any, Dict, Iterable, List, Optional

from together.types.common import FinishReason, ObjectType 
from together.types.chat_completions import ChatCompletionResponse, MessageRole, UsageData, ChatCompletionChoicesData, ChatCompletionMessage

from ..logger import GenerationRequestMessage, Generation
from ..components.attachment import UrlAttachment
from ...scribe import scribe


class TogetherUtils:
    """Utility class for Together AI API integration with Maxim.

    This class provides static utility methods for parsing and processing
    Together AI API requests and responses to integrate with Maxim's logging
    and monitoring system. It handles message format conversion, parameter
    extraction, response standardization, and image attachment processing.

    All methods are static and can be called directly on the class without
    instantiation. The class follows the same patterns as other provider
    integrations in the Maxim SDK.
    """

    @staticmethod
    def parse_message_param(
        messages: Iterable[Dict[str, Any]],
        override_role: Optional[str] = None,
    ) -> List[GenerationRequestMessage]:
        """Parse Together AI message parameters into Maxim format.

        This method converts Together AI message dictionaries into Maxim's
        GenerationRequestMessage format for consistent logging and tracking.
        It handles various message formats including string content and
        structured content blocks with multimodal support.

        Args:
            messages (Iterable[Dict[str, Any]]): Iterable of Together AI message
                dictionaries to be parsed. Each message should have 'role' and
                'content' keys following Together AI's message format.
            override_role (Optional[str]): Optional role to override the message role.
                If provided, all messages will use this role instead of their
                original role. Defaults to None.

        Returns:
            List[GenerationRequestMessage]: List of parsed messages in Maxim format,
                with role and content extracted and standardized.
        """
        parsed_messages: List[GenerationRequestMessage] = []

        for msg in messages:
            role = override_role or msg.get("role", "user")
            content = msg.get("content", "")

            if isinstance(content, list):
                # Handle content blocks for multimodal messages
                text_content = ""
                for block in content:
                    if isinstance(block, dict) and block.get("type") == "text":
                        text_content += block.get("text", "")
                parsed_messages.append(
                    GenerationRequestMessage(role=role, content=text_content)
                )
            else:
                # Handle simple string content
                parsed_messages.append(
                    GenerationRequestMessage(role=role, content=str(content))
                )

        return parsed_messages

    @staticmethod
    def get_model_params(**kwargs: Any) -> Dict[str, Any]:
        """Extract and normalize model parameters for Maxim logging.

        This method extracts relevant model parameters from Together AI API
        calls and formats them for consistent logging in Maxim. It handles
        common parameters like temperature, max_tokens, and Together AI specific
        parameters while filtering out internal parameters.

        Args:
            **kwargs (Any): Keyword arguments that may contain model parameters
                from Together AI API calls. Can include parameters like temperature,
                max_tokens, top_p, frequency_penalty, etc.

        Returns:
            Dict[str, Any]: Dictionary containing normalized model parameters
                with non-None values only. Internal parameters like 'messages'
                and 'model' are excluded from the result.
        """
        model_params = {}
        skip_keys = ["messages", "model"]

        # Common parameters that Together AI supports
        param_keys = [
            "temperature",
            "top_p",
            "max_tokens",
            "frequency_penalty",
            "presence_penalty",
            "repetition_penalty",
            "top_k",
            "min_p",
            "safety_model",
            "response_format",
            "tools",
            "tool_choice",
            "stream",
            "logprobs",
            "echo",
            "n",
            "stop",
        ]

        # Add explicitly known parameters
        for key in param_keys:
            if key in kwargs and kwargs[key] is not None and key not in skip_keys:
                model_params[key] = kwargs[key]

        # Add any other parameters that aren't in skip_keys
        for key, value in kwargs.items():
            if key not in param_keys and key not in skip_keys and value is not None:
                model_params[key] = value

        return model_params

    @staticmethod
    def parse_chunks_to_response(
        content: str, usage_data: Optional[UsageData] = None
    ) -> ChatCompletionResponse:
        """Create a response object from streaming chunks for parsing.
        
        This method constructs a response object compatible with the parse_completion
        method from accumulated streaming content and usage data. It creates a
        structured response that mimics the Together AI response format.
        
        Args:
            content (str): The accumulated content from streaming chunks that
                represents the complete response text.
            usage_data (UsageData): Usage information from the final chunk
                containing token counts and other usage metrics.
            
        Returns:
            Response: Response object with proper attributes that can be processed
                by parse_completion() method. Contains choices, usage, and
                standard response metadata.
        """
       
        return ChatCompletionResponse(
            id=f"streaming-response-{uuid.uuid4()}",
            created=int(time.time()),
            choices=[ChatCompletionChoicesData(
                index=0,
                message=ChatCompletionMessage(role=MessageRole.ASSISTANT, content=content),
                finish_reason=FinishReason.StopSequence
            )],
            usage=usage_data,
            object=ObjectType.ChatCompletion,
        )

    @staticmethod
    def parse_completion(completion: ChatCompletionResponse) -> Dict[str, Any]:
        """Parse Together AI completion response into standardized format.

        This method converts a Together AI ChatCompletionResponse object into
        a standardized dictionary format that can be consistently processed
        by Maxim's logging system. It handles various response formats and
        provides fallback parsing for edge cases.

        Args:
            completion (ChatCompletionResponse): The completion response object
                from Together AI API containing the generated content, choices,
                usage information, and other response metadata.

        Returns:
            Dict[str, Any]: Standardized response dictionary with the following structure:
                - id: Unique identifier for the response
                - created: Unix timestamp of creation
                - choices: List of choices with message content and metadata
                - usage: Token usage statistics (prompt, completion, total)
        """
        # Handle Together ChatCompletionResponse format
        if hasattr(completion, 'choices') and hasattr(completion, 'id'):
            parsed_response = {
                "id": completion.id,
                "created": getattr(completion, 'created', int(time.time())),
                "choices": [],
            }

            for choice in completion.choices or []:
                choice_data = {
                    "index": getattr(choice, 'index', 0),
                    "message": {
                        "role": getattr(choice.message, 'role', 'assistant'),
                        "content": getattr(choice.message, 'content', ''),
                    },
                    "finish_reason": getattr(choice, 'finish_reason', None),
                }

                # Add tool calls if present
                if (
                    choice.message is not None
                    and hasattr(choice.message, "tool_calls")
                    and choice.message.tool_calls
                ):
                    choice_data["message"]["tool_calls"] = choice.message.tool_calls

                parsed_response["choices"].append(choice_data)

            # Add usage information if available
            if (
                hasattr(completion, "usage")
                and completion.usage is not None
                and completion.usage.prompt_tokens is not None
                and completion.usage.completion_tokens is not None
                and completion.usage.total_tokens is not None
            ):
                parsed_response["usage"] = {
                    "prompt_tokens": getattr(completion.usage, 'prompt_tokens', 0),
                    "completion_tokens": getattr(completion.usage, 'completion_tokens', 0),
                    "total_tokens": getattr(completion.usage, 'total_tokens', 0),
                }

            return parsed_response

        # Fallback for dict-like responses
        elif isinstance(completion, dict):
            return completion

        return {}

    @staticmethod
    def add_image_attachments_from_messages(generation: Generation, messages: Iterable[Dict[str, Any]]) -> None:
        """Extract image URLs from messages and add them as attachments to the generation.

        This method scans through Together AI messages to find image URLs in content
        blocks and automatically adds them as URL attachments to the generation object.
        It handles the multimodal message format where images are embedded within
        content arrays.

        Args:
            generation (Generation): The Maxim generation object to add attachments to.
                If None, the method will return early without processing.
            messages (Iterable[Dict[str, Any]]): The messages to scan for image URLs.
                Should follow Together AI's message format with content arrays
                containing image_url objects.

        Returns:
            None: This method modifies the generation object in-place by adding
                attachments and does not return any value.
        """
        if generation is None or not messages:
            return

        try:
            for message in messages:
                if isinstance(message, dict) and message.get("role") == "user":
                    content = message.get("content", [])
                    if isinstance(content, list):
                        for content_item in content:
                            if isinstance(content_item, dict) and content_item.get("type") == "image_url":
                                image_url_data = content_item.get("image_url", {})
                                image_url = image_url_data.get("url", "")
                                if image_url:
                                    generation.add_attachment(UrlAttachment(
                                        url=image_url,
                                        name="User Image",
                                        mime_type="image"
                                    ))
        except Exception as e:
            generation.error({"message": str(e)})
            scribe().warning(f"[MaximSDK] Error adding image attachments: {e}")



import asyncio
import math
import re
import threading
import time
from dataclasses import dataclass
from typing import Any, Awaitable, Callable, Dict, Generic, List, Optional, Union, final

from ..apis import MaximAPI
from ..dataset import sanitize_data_structure
from ..evaluators import BaseEvaluator
from ..models import (
    DatasetRow,
    Evaluator,
    EvaluatorType,
    HumanEvaluationConfig,
    RunResult,
    RunStatus,
    RunType,
    T,
    TestRun,
    TestRunConfig,
    TestRunEntry,
    TestRunLogger,
    TestRunStatus,
    TestRunWithDatasetEntry,
    YieldedOutput,
    YieldedOutputMeta,
)
from ..models.dataset import Data, LocalData
from ..models.evaluator import (
    LocalEvaluationResultWithId,
    LocalEvaluatorResultParameter,
)
from ..models.test_run import (
    PromptChainVersionConfig,
    PromptVersionConfig,
    WorkflowConfig,
    YieldedOutputTokenUsage,
)
from ..test_runs.run_utils import (
    get_input_expected_output_and_context_from_row,
    get_variables_from_row,
    process_awaitable,
    run_local_evaluations,
)
from ..test_runs.sanitization_utils import sanitize_data, sanitize_evaluators
from ..test_runs.utils import (
    EvaluatorNameToIdAndPassFailCriteria,
    get_evaluator_config_from_evaluator_name_and_pass_fail_criteria,
    get_local_evaluator_name_to_id_and_pass_fail_criteria_map,
)
from ..utils import Semaphore


def calculate_polling_interval(
    timeout_minutes: float, is_ai_evaluator_in_use: bool = False
) -> int:
    points = [
        (10, 5),
        (15, 5),
        (30, 10),
        (60, 15),
        (120, 30),
        (1440, 120),
    ]

    lower_point = points[0]
    upper_point = points[-1]
    for i in range(len(points) - 1):
        if points[i][0] <= timeout_minutes <= points[i + 1][0]:
            lower_point = points[i]
            upper_point = points[i + 1]
            break

    x1, y1 = lower_point
    x2, y2 = upper_point
    if x1 == x2:
        return y1

    t = (timeout_minutes - x1) / (x2 - x1)
    p = 2
    interpolated_value = y1 + (y2 - y1) * pow(t, p)

    return min(max(round(interpolated_value), 15 if is_ai_evaluator_in_use else 5), 120)


def get_all_keys_by_value(obj: Optional[dict[Any, Any]], value: Any) -> List[str]:
    if obj is None:
        return []
    return [key for key, val in obj.items() if val == value]


@dataclass
class ProcessedEntry:
    entry: TestRunEntry
    meta: Optional[YieldedOutputMeta] = None

    def to_dict(self) -> Dict[str, Any]:
        return {
            "entry": self.entry.to_dict(),
            "meta": self.meta.to_dict() if self.meta else None,
        }


@final
class TestRunBuilder(Generic[T]):
    """
    Builder for test runs.
    """

    def __init__(
        self,
        base_url: str,
        api_key: str,
        name: str,
        workspace_id: str,
        evaluators: List[Union[str, BaseEvaluator]],
    ):
        """
        Constructor
        """
        self._config = TestRunConfig(
            base_url=base_url,
            api_key=api_key,
            name=name,
            evaluators=evaluators,
            in_workspace_id=workspace_id,
        )
        self._maxim_apis = MaximAPI(base_url, api_key)

    def __process_entry(
        self,
        index: int,
        input: Optional[str],
        expected_output: Optional[str],
        context_to_evaluate: Optional[Union[str, List[str]]],
        output_function: Optional[
            Callable[[LocalData], Union[YieldedOutput, Awaitable[YieldedOutput]]]
        ],
        get_row: Callable[[int], Optional[LocalData]],
        logger: TestRunLogger,
        evaluator_name_to_id_and_pass_fail_criteria_map: Dict[
            str, EvaluatorNameToIdAndPassFailCriteria
        ],
    ) -> ProcessedEntry:
        """
        Process a single test run entry

        Args:
            index (int): The index of the entry
            keys (dict[str, Optional[str]]): Mapping of column names to keys in the data
            output_function (Callable[[dict[str, Any]], YieldedOutput]): Function to generate output
            get_row (Callable[[int], Optional[dict[str, Any]]]): Function to retrieve a row from the dataset
            logger (TestRunLogger): Logger instance

        Returns:
            ProcessedEntry: Contains processed entry and metadata
        """
        row = get_row(index)
        if row is None:
            raise ValueError(f"Dataset entry {index} is missing")

        output: Optional[Union[Awaitable[YieldedOutput], YieldedOutput]]
        if output_function is not None:
            output = output_function(row)
        elif self._config.workflow is not None:
            workflow_output = self._maxim_apis.execute_workflow_for_data(
                self._config.workflow.id, row, self._config.workflow.context_to_evaluate
            )
            output = YieldedOutput(
                data=(
                    workflow_output.output if workflow_output.output is not None else ""
                ),
                retrieved_context_to_evaluate=workflow_output.context_to_evaluate,
                meta=YieldedOutputMeta(
                    entity_type="WORKFLOW",
                    entity_id=self._config.workflow.id,
                    usage=YieldedOutputTokenUsage(
                        latency=workflow_output.latency,
                        completion_tokens=0,
                        prompt_tokens=0,
                        total_tokens=0,
                    ),
                ),
            )
        elif self._config.prompt_version is not None:
            variables = (
                get_variables_from_row(row, self._config.data_structure)
                if self._config.data_structure
                else {}
            )
            prompt_output = self._maxim_apis.execute_prompt_for_data(
                self._config.prompt_version.id,
                input if input is not None else "",
                variables,
                self._config.prompt_version.context_to_evaluate,
            )
            output = YieldedOutput(
                data=prompt_output.output if prompt_output.output is not None else "",
                retrieved_context_to_evaluate=prompt_output.context_to_evaluate,
                meta=YieldedOutputMeta(
                    entity_type="PROMPT",
                    entity_id=self._config.prompt_version.id,
                    usage=prompt_output.usage,
                    cost=prompt_output.cost,
                ),
            )
        elif self._config.prompt_chain_version is not None:
            variables = (
                get_variables_from_row(row, self._config.data_structure)
                if self._config.data_structure
                else {}
            )
            prompt_chain_output = self._maxim_apis.execute_prompt_chain_for_data(
                self._config.prompt_chain_version.id,
                input if input is not None else "",
                variables,
                self._config.prompt_chain_version.context_to_evaluate,
            )
            output = YieldedOutput(
                data=(
                    prompt_chain_output.output
                    if prompt_chain_output.output is not None
                    else ""
                ),
                retrieved_context_to_evaluate=prompt_chain_output.context_to_evaluate,
                meta=YieldedOutputMeta(
                    entity_type="PROMPT_CHAIN",
                    entity_id=self._config.prompt_chain_version.id,
                    usage=prompt_chain_output.usage,
                    cost=prompt_chain_output.cost,
                ),
            )
        else:
            raise ValueError(
                "Found no output function to execute, please make sure you have either `yields_output`, `with_prompt_version_id`, `with_prompt_chain_version_id` or `with_workflow_id` set."
            )

        yielded_output: YieldedOutput

        if isinstance(output, Awaitable):
            yielded_output = asyncio.run(process_awaitable(output))
        else:
            yielded_output = output

        if yielded_output is not None:
            if (
                yielded_output.retrieved_context_to_evaluate is not None
                and context_to_evaluate is not None
            ):
                logger.info(
                    "Overriding context_to_evaluate from output over dataset entry"
                )
            context_to_evaluate = yielded_output.retrieved_context_to_evaluate

        local_evaluators: List[BaseEvaluator] = []
        for evaluator in self._config.evaluators:
            if isinstance(evaluator, BaseEvaluator):
                local_evaluators.append(evaluator)
            else:
                continue

        local_evaluation_results_awaitable = run_local_evaluations(
            local_evaluators,
            row,
            LocalEvaluatorResultParameter(
                output=yielded_output.data if yielded_output is not None else "",
                context_to_evaluate=context_to_evaluate,
            ),
        )
        local_evaluation_results = asyncio.run(
            process_awaitable(local_evaluation_results_awaitable)
        )

        local_evaluation_results_with_ids: List[LocalEvaluationResultWithId] = []
        for local_evaluation_result in local_evaluation_results:
            local_evaluation_results_with_ids.append(
                LocalEvaluationResultWithId(
                    result=local_evaluation_result.result,
                    id=evaluator_name_to_id_and_pass_fail_criteria_map[
                        local_evaluation_result.name
                    ].id,
                    name=local_evaluation_result.name,
                    pass_fail_criteria=local_evaluation_result.pass_fail_criteria,
                )
            )

        return ProcessedEntry(
            entry=TestRunEntry(
                output=yielded_output.data if yielded_output is not None else None,
                input=input,
                expected_output=expected_output,
                context_to_evaluate=context_to_evaluate,
                variables=(
                    get_variables_from_row(row, self._config.data_structure)
                    if self._config.data_structure
                    else {}
                ),
                local_evaluation_results=local_evaluation_results_with_ids,
            ),
            meta=yielded_output.meta if yielded_output is not None else None,
        )

    def with_data_structure(self, data: T) -> "TestRunBuilder[T]":
        """
        Set the data structure for the test run

        Args:
            data (T): The data structure to use

        Returns:
            TestRunBuilder[T]: The current TestRunBuilder instance for method chaining
        """
        sanitize_data_structure(data)
        self._config.data_structure = data
        return self

    def with_data(self, data: Data) -> "TestRunBuilder[T]":
        """
        Set the data for the test run

        Args:
            data (DataValue[T]): The data to use

        Returns:
            TestRunBuilder[T]: The current TestRunBuilder instance for method chaining
        """
        if self._config.data_structure is not None:
            sanitize_data(data, self._config.data_structure)
        self._config.data = data
        return self

    def with_evaluators(
        self, *evaluators: Union[str, BaseEvaluator]
    ) -> "TestRunBuilder[T]":
        """
        Add evaluators to the test run

        Args:
            *evaluators (str): The evaluators to add

        Returns:
            TestRunBuilder[T]: The current TestRunBuilder instance for method chaining
        """
        evaluators_list: List[Union[BaseEvaluator, str]] = []
        for evaluator in evaluators:
            evaluators_list.append(evaluator)
        sanitize_evaluators(evaluators_list)
        self._config.evaluators = evaluators_list
        return self

    def with_human_evaluation_config(
        self, config: HumanEvaluationConfig
    ) -> "TestRunBuilder[T]":
        """
        Set the human evaluation configuration for the test run

        Args:
            config (HumanEvaluationConfig): The human evaluation configuration to use

        Returns:
            TestRunBuilder[T]: The current TestRunBuilder instance for method chaining
        """
        email_regex = re.compile(
            r"^(?!\.)(?!.*\.\.)([A-Z0-9_\'+\-\.]*)[A-Z0-9_+-]@([A-Z0-9][A-Z0-9\-]*\.)+[A-Z]{2,}$",
            re.IGNORECASE,
        )
        invalid_emails = [
            email for email in config.emails if not email_regex.match(email)
        ]
        if len(invalid_emails) > 0:
            raise ValueError(f"Invalid email addresses: {', '.join(invalid_emails)}")
        self._config.human_evaluation_config = config
        return self

    def with_workflow_id(
        self, workflow_id: str, context_to_evaluate: Optional[str] = None
    ) -> "TestRunBuilder[T]":
        """
        Set the workflow ID for the test run. Optionally, you can also set the context to evaluate for the workflow. (Note: setting the context to evaluate will end up overriding the CONTEXT_TO_EVALUATE dataset column value)

        Args:
            workflow_id (str): The ID of the workflow to use
            context_to_evaluate (Optional[str]): The context to evaluate for the workflow (variable name essentially).

        Returns:
            TestRunBuilder[T]: The current TestRunBuilder instance for method chaining

        Raises:
            ValueError: If a prompt version ID, prompt chain version ID or output function is already set for this run builder
        """
        if self._config.prompt_version is not None:
            raise ValueError(
                "Prompt version id is already set for this run builder. You can use either one of with_prompt_version_id, with_prompt_chain_version_id, with_workflow_id or yields_output in a test run."
            )
        if self._config.prompt_chain_version is not None:
            raise ValueError(
                "Prompt chain version id is already set for this run builder. You can use either one of with_prompt_version_id, with_prompt_chain_version_id, with_workflow_id or yields_output in a test run."
            )
        if self._config.output_function is not None:
            raise ValueError(
                "yeilds_output is already set for this run builder. You can use either one of with_prompt_version_id, with_prompt_chain_version_id, with_workflow_id or yields_output in a test run."
            )
        self._config.workflow = WorkflowConfig(
            id=workflow_id,
            context_to_evaluate=context_to_evaluate,
        )
        return self

    def with_prompt_version_id(
        self, prompt_version_id: str, context_to_evaluate: Optional[str] = None
    ) -> "TestRunBuilder[T]":
        """
        Set the prompt version ID for the test run. Optionally, you can also set the context to evaluate for the prompt. (Note: setting the context to evaluate will end up overriding the CONTEXT_TO_EVALUATE dataset column value)

        Args:
            prompt_version_id (str): The ID of the prompt version to use
            context_to_evaluate (Optional[str]): The context to evaluate for the prompt (variable name essentially).

        Returns:
            TestRunBuilder[T]: The current TestRunBuilder instance for method chaining

        Raises:
            ValueError: If a workflow ID, prompt chain version ID or output function is already set for this run builder
        """
        if self._config.workflow is not None:
            raise ValueError(
                "Workflow id is already set for this run builder. You can use either one of with_prompt_version_id, with_prompt_chain_version_id, with_workflow_id or yields_output in a test run."
            )
        if self._config.prompt_chain_version is not None:
            raise ValueError(
                "Prompt chain version id is already set for this run builder. You can use either one of with_prompt_version_id, with_prompt_chain_version_id, with_workflow_id or yields_output in a test run."
            )
        if self._config.output_function is not None:
            raise ValueError(
                "yields_output is already set for this run builder. You can use either one of with_prompt_version_id, with_prompt_chain_version_id, with_workflow_id or yields_output in a test run."
            )
        self._config.prompt_version = PromptVersionConfig(
            id=prompt_version_id,
            context_to_evaluate=context_to_evaluate,
        )
        return self

    def with_prompt_chain_version_id(
        self, prompt_chain_version_id: str, context_to_evaluate: Optional[str] = None
    ) -> "TestRunBuilder[T]":
        """
        Set the prompt chain version ID for the test run. Optionally, you can also set the context to evaluate for the prompt chain. (Note: setting the context to evaluate will end up overriding the CONTEXT_TO_EVALUATE dataset column value)

        Args:
            prompt_chain_version_id (str): The ID of the prompt chain version to use
            context_to_evaluate (Optional[str]): The context to evaluate for the prompt chain (variable name essentially).

        Returns:
            TestRunBuilder[T]: The current TestRunBuilder instance for method chaining

        Raises:
            ValueError: If a workflow ID, prompt version ID or output function is already set for this run builder
        """
        if self._config.workflow is not None:
            raise ValueError(
                "Workflow id is already set for this run builder. You can use either one of with_prompt_version_id, with_prompt_chain_version_id, with_workflow_id or yields_output in a test run."
            )
        if self._config.prompt_version is not None:
            raise ValueError(
                "Prompt version id is already set for this run builder. You can use either one of with_prompt_version_id, with_prompt_chain_version_id, with_workflow_id or yields_output in a test run."
            )
        if self._config.output_function is not None:
            raise ValueError(
                "yields_output is already set for this run builder. You can use either one of with_prompt_version_id, with_prompt_chain_version_id, with_workflow_id or yields_output in a test run."
            )
        self._config.prompt_chain_version = PromptChainVersionConfig(
            id=prompt_chain_version_id,
            context_to_evaluate=context_to_evaluate,
        )
        return self

    def yields_output(
        self,
        output_function: Callable[
            [LocalData], Union[YieldedOutput, Awaitable[YieldedOutput]]
        ],
    ) -> "TestRunBuilder[T]":
        """
        Set the output function for the test run

        Args:
            output_function (Callable[[T], Union[YieldedOutput, Awaitable[YieldedOutput]]]): The output function to use

        Returns:
            TestRunBuilder[T]: The current TestRunBuilder instance for method chaining

        Raises:
            ValueError: If a workflow ID, prompt chain version ID or prompt version ID is already set for this run builder
        """
        if self._config.workflow is not None:
            raise ValueError(
                "Workflow id is already set for this run builder. You can use either one of with_prompt_version_id, with_prompt_chain_version_id, with_workflow_id or yields_output in a test run."
            )
        if self._config.prompt_chain_version is not None:
            raise ValueError(
                "Prompt chain version id is already set for this run builder. You can use either one of with_prompt_version_id, with_prompt_chain_version_id, with_workflow_id or yields_output in a test run."
            )
        if self._config.prompt_version is not None:
            raise ValueError(
                "Prompt version id is already set for this run builder. You can use either one of with_prompt_version_id, prompt_chain_version_id, with_workflow_id or yields_output in a test run."
            )
        self._config.output_function = output_function
        return self

    def with_concurrency(self, concurrency: int) -> "TestRunBuilder[T]":
        """
        Set the concurrency level for the test run

        Args:
            concurrency (int): The concurrency level to use

        Returns:
            TestRunBuilder[T]: The current TestRunBuilder instance for method chaining
        """
        self._config.concurrency = concurrency
        return self

    def with_logger(self, logger: TestRunLogger) -> "TestRunBuilder[T]":
        """
        Set the logger for the test run

        Args:
            logger (TestRunLogger): The logger to use

        Returns:
            TestRunBuilder[T]: The current TestRunBuilder instance for method chaining
        """
        self._config.logger = logger
        return self

    def _run_test_with_local_data(
        self,
        test_run: TestRun,
        get_row: Callable[[int], Optional[LocalData]],
        on_entry_failed: Callable[[int], None],
        on_dataset_finished: Callable[[], None],
        evaluator_name_to_id_and_pass_fail_criteria_map: Dict[
            str, EvaluatorNameToIdAndPassFailCriteria
        ],
    ):
        """
        Run the test with local data

        Args:
            test_run (TestRun): The test run to execute
            get_row (Callable[[int], Optional[Dict[str, Any]]]): Function to retrieve a row from the dataset
            on_entry_failed (Callable[[int], None]): Callback for when an entry fails
            on_dataset_finished (Callable[[], None]): Callback for when the dataset is finished
        """
        semaphore = Semaphore.get(
            f"test_run:{test_run.workspace_id}:{test_run.id}",
            self._config.concurrency or 10,
        )
        data_structure = self._config.data_structure
        try:
            input_key = get_all_keys_by_value(data_structure, "INPUT")[0]
        except IndexError:
            input_key = None
        try:
            expectedOutputKey = get_all_keys_by_value(
                data_structure, "EXPECTED_OUTPUT"
            )[0]
        except IndexError:
            expectedOutputKey = None
        try:
            contextToEvaluateKey = get_all_keys_by_value(
                data_structure, "CONTEXT_TO_EVALUATE"
            )[0]
        except IndexError:
            contextToEvaluateKey = None

        def process_row(
            index: int,
            row: LocalData,
            evaluator_name_to_id_and_pass_fail_criteria_map: Dict[
                str, EvaluatorNameToIdAndPassFailCriteria
            ],
        ) -> None:
            try:
                if row is None:
                    raise ValueError(f"Dataset entry {index} is missing")
                input, expected_output, context_to_evaluate = (
                    get_input_expected_output_and_context_from_row(
                        input_key, expectedOutputKey, contextToEvaluateKey, row
                    )
                )
                if (
                    any(isinstance(e, BaseEvaluator) for e in self._config.evaluators)
                    or self._config.output_function is not None
                ):
                    result = self.__process_entry(
                        index=index,
                        input=input,
                        expected_output=expected_output,
                        context_to_evaluate=context_to_evaluate,
                        output_function=self._config.output_function,
                        get_row=lambda index: row,
                        logger=self._config.logger,
                        evaluator_name_to_id_and_pass_fail_criteria_map=evaluator_name_to_id_and_pass_fail_criteria_map,
                    )
                    # pushing this entry to test run
                    self._maxim_apis.push_test_run_entry(
                        test_run=test_run,
                        entry=result.entry,
                        run_config=(
                            {
                                "cost": (
                                    result.meta.cost.to_dict()
                                    if result.meta.cost is not None
                                    else None
                                ),
                                "usage": (
                                    result.meta.usage.to_dict()
                                    if result.meta.usage is not None
                                    else None
                                ),
                            }
                            if result.meta is not None
                            else None
                        ),
                    )
                else:
                    # pushing directly
                    self._maxim_apis.push_test_run_entry(
                        test_run=test_run,
                        entry=TestRunEntry(
                            variables=(
                                get_variables_from_row(row, self._config.data_structure)
                                if self._config.data_structure
                                else {}
                            ),
                            input=input,
                            expected_output=expected_output,
                            context_to_evaluate=context_to_evaluate,
                        ),
                    )
            except Exception as e:
                self._config.logger.error(
                    f"Error while running data entry at index [{index}]: {str(e)}"
                )
                on_entry_failed(index)

        def process_all_entries() -> None:
            threads = []
            index = 0
            while True:
                try:
                    semaphore.acquire()
                    # getting the entry
                    row = get_row(index)
                    if row is None:
                        on_dataset_finished()
                        break
                    # sanitizing data
                    try:
                        if self._config.data_structure is None:
                            raise ValueError(
                                "Data structure is required to run a test with local data as a function"
                            )
                        sanitize_data(row, self._config.data_structure)
                    except ValueError as e:
                        self._config.logger.error(
                            f"Invalid data entry at index [{index}]: {str(e)}"
                        )
                        on_entry_failed(index)
                        continue
                    index += 1
                    thread = threading.Thread(
                        target=process_row,
                        args=(
                            index,
                            row,
                            evaluator_name_to_id_and_pass_fail_criteria_map,
                        ),
                    )
                    thread.start()
                    threads.append(thread)
                except Exception as e:
                    self._config.logger.error(
                        f"Error while running data entry at index [{index}]: {str(e)}",
                    )
                    on_entry_failed(index)
                finally:
                    semaphore.release()

        thread = threading.Thread(target=process_all_entries, args=())
        thread.start()

    def _run_test_with_dataset_id(
        self,
        test_run: TestRun,
        dataset_id: str,
        on_entry_failed: Callable[[int], None],
        on_dataset_finished: Callable[[], None],
        evaluator_name_to_id_and_pass_fail_criteria_map: Dict[
            str, EvaluatorNameToIdAndPassFailCriteria
        ],
    ) -> None:
        """
        Run the test with a dataset ID

        Args:
            test_run (TestRun): The test run to execute
            dataset_id (str): The ID of the dataset to use
            on_entry_failed (Callable[[int], None]): Callback for when an entry fails
            on_dataset_finished (Callable[[], None]): Callback for when the dataset is finished
        """
        semaphore = Semaphore.get(
            f"test_run:{test_run.workspace_id}:{test_run.id}",
            self._config.concurrency or 10,
        )
        data_structure = self._maxim_apis.get_dataset_structure(dataset_id)
        self._maxim_apis.attach_dataset_to_test_run(
            test_run_id=test_run.id, dataset_id=dataset_id
        )
        try:
            input_key = get_all_keys_by_value(data_structure, "INPUT")[0]
        except IndexError:
            input_key = None
        try:
            expectedOutputKey = get_all_keys_by_value(
                data_structure, "EXPECTED_OUTPUT"
            )[0]
        except IndexError:
            expectedOutputKey = None
        try:
            contextToEvaluateKey = get_all_keys_by_value(
                data_structure, "CONTEXT_TO_EVALUATE"
            )[0]
        except IndexError:
            contextToEvaluateKey = None

        def process_dataset_entry(
            index: int,
            row: DatasetRow,
            dataset_id: str,
            evaluator_name_to_id_and_pass_fail_criteria_map: Dict[
                str, EvaluatorNameToIdAndPassFailCriteria
            ],
        ) -> None:
            try:
                row_data: LocalData = row.to_dict()["data"]
                if row_data is None:
                    raise ValueError(f"Dataset entry {index} is missing")
                input, expected_output, context_to_evaluate = (
                    get_input_expected_output_and_context_from_row(
                        input_key, expectedOutputKey, contextToEvaluateKey, row_data
                    )
                )

                if (
                    any(isinstance(e, BaseEvaluator) for e in self._config.evaluators)
                    or self._config.output_function is not None
                ):
                    # processing the entry
                    result = self.__process_entry(
                        index=index,
                        input=input,
                        expected_output=expected_output,
                        context_to_evaluate=context_to_evaluate,
                        output_function=self._config.output_function,
                        get_row=lambda index: row.to_dict()["data"],
                        logger=self._config.logger,
                        evaluator_name_to_id_and_pass_fail_criteria_map=evaluator_name_to_id_and_pass_fail_criteria_map,
                    )
                    # pushing this entry to test run
                    self._maxim_apis.push_test_run_entry(
                        test_run=TestRunWithDatasetEntry(
                            test_run=test_run,
                            dataset_id=dataset_id,
                            dataset_entry_id=row.id,
                        ),
                        entry=result.entry,
                        run_config=(
                            {
                                "cost": (
                                    result.meta.cost.to_dict()
                                    if result.meta.cost is not None
                                    else None
                                ),
                                "usage": (
                                    result.meta.usage.to_dict()
                                    if result.meta.usage is not None
                                    else None
                                ),
                            }
                            if result.meta
                            else None
                        ),
                    )
                else:
                    # pushing directly
                    self._maxim_apis.push_test_run_entry(
                        test_run=TestRunWithDatasetEntry(
                            test_run=test_run,
                            dataset_id=dataset_id,
                            dataset_entry_id=row.id,
                        ),
                        entry=TestRunEntry(
                            variables=(
                                get_variables_from_row(row_data, data_structure)
                                if data_structure
                                else {}
                            ),
                            input=input,
                            expected_output=expected_output,
                            context_to_evaluate=context_to_evaluate,
                        ),
                    )
            except Exception as e:
                self._config.logger.error(
                    f"Error while running data entry at index [{index}]: {str(e)}",
                )
                on_entry_failed(index)
                raise e

        def process_all_dataset_entries(dataset_id: str) -> None:
            threads = []
            index = 0
            total_rows = self._maxim_apis.get_dataset_total_rows(dataset_id)
            for index in range(total_rows):
                try:
                    semaphore.acquire()
                    # getting the entry
                    row = self._maxim_apis.get_dataset_row(dataset_id, index)
                    if row is None:
                        break
                    thread = threading.Thread(
                        target=process_dataset_entry,
                        args=(
                            index,
                            row,
                            dataset_id,
                            evaluator_name_to_id_and_pass_fail_criteria_map,
                        ),
                    )
                    thread.start()
                    threads.append(thread)
                except Exception as e:
                    self._config.logger.error(
                        f"Error while running data entry at index [{index}]: {str(e)}"
                    )
                    on_entry_failed(index)
                finally:
                    semaphore.release()
            on_dataset_finished()

            for thread in threads:
                thread.join()

        thread = threading.Thread(
            target=process_all_dataset_entries, args=(dataset_id,)
        )
        thread.start()

    def run(self, timeout_in_minutes: Optional[int] = 10) -> Optional[RunResult]:
        """
        Run the test

        Args:
            timeout_in_minutes (Optional[int]): The timeout in minutes. Defaults to 10.

        Returns:
            RunResult: The result of the test run
        """
        try:
            errors: list[str] = []
            self._config.logger.info(message="Validating test run config...")
            if self._config.name == "":
                errors.append("Name is required to run a test.")
            if self._config.in_workspace_id == "":
                errors.append("Workspace id is required to run a test.")
            if (
                self._config.output_function is None
                and self._config.workflow is None
                and self._config.prompt_version is None
                and self._config.prompt_chain_version is None
            ):
                errors.append(
                    "One of output function (by calling yields_output) or workflow id (by calling with_workflow_id) or prompt version id (by calling with_prompt_version_id) or prompt chain version id (by calling with_prompt_chain_version_id) is required to run a test."
                )
            if self._config.data is None:
                errors.append("Dataset id is required to run a test.")
            if len(errors) > 0:
                raise ValueError(
                    "Missing required configuration for test\n" + "\n".join(errors)
                )
            self._config.logger.info(message="Sanitizing data...")
            sanitize_data_structure(self._config.data_structure)
            if isinstance(self._config.data, List):
                if self._config.data_structure:
                    sanitize_data(self._config.data, self._config.data_structure)
            self._config.logger.info(message="Sanitizing evaluators...")
            sanitize_evaluators(self._config.evaluators)
            evaluator_configs: List[Evaluator] = []
            evaluator_name_to_id_and_pass_fail_criteria_map = (
                get_local_evaluator_name_to_id_and_pass_fail_criteria_map(
                    self._config.evaluators
                )
            )
            for evaluator in self._config.evaluators or []:
                if isinstance(evaluator, str):
                    try:
                        self._config.logger.info(
                            message=f"Verifying if {evaluator} is added to the workspace.."
                        )
                        evaluator_config = self._maxim_apis.fetch_platform_evaluator(
                            name=evaluator, in_workspace_id=self._config.in_workspace_id
                        )
                        evaluator_configs.append(evaluator_config)
                    except Exception as e:
                        raise ValueError(
                            f"Failed to fetch evaluator {evaluator}"
                        ) from e
                else:
                    for name in evaluator.names:
                        evaluator_config = get_evaluator_config_from_evaluator_name_and_pass_fail_criteria(
                            id=evaluator_name_to_id_and_pass_fail_criteria_map[name].id,
                            name=name,
                            pass_fail_criteria=evaluator_name_to_id_and_pass_fail_criteria_map[
                                name
                            ].pass_fail_criteria,
                        )
                        evaluator_configs.append(evaluator_config)

            if any(
                evaluator.type.value == EvaluatorType.HUMAN.value
                for evaluator in evaluator_configs
            ):
                if self._config.human_evaluation_config is None:
                    raise ValueError(
                        "Human evaluator found in evaluators, but no human evaluation config was provided."
                    )

            name = self._config.name
            data = self._config.data
            workspace_id = self._config.in_workspace_id
            human_evaluation_config = self._config.human_evaluation_config
            failed_entry_indices = []
            all_entries_processed = threading.Event()

            def mark_all_entries_processed() -> None:
                nonlocal all_entries_processed
                all_entries_processed.set()

            try:
                self._config.logger.info(f"Creating test run: {name}")
                requires_local_run = False
                if (
                    any(isinstance(e, BaseEvaluator) for e in self._config.evaluators)
                    or self._config.output_function is not None
                ):
                    requires_local_run = True
                test_run = self._maxim_apis.create_test_run(
                    name=name,
                    workspace_id=workspace_id,
                    run_type=RunType.SINGLE,
                    workflow_id=(
                        self._config.workflow.id
                        if self._config.workflow is not None
                        else None
                    ),
                    prompt_version_id=(
                        self._config.prompt_version.id
                        if self._config.prompt_version is not None
                        else None
                    ),
                    prompt_chain_version_id=(
                        self._config.prompt_chain_version.id
                        if self._config.prompt_chain_version is not None
                        else None
                    ),
                    evaluator_config=evaluator_configs,
                    human_evaluation_config=human_evaluation_config or None,
                    requires_local_run=requires_local_run,
                )
                try:
                    if data is not None:
                        if isinstance(data, str):
                            self._run_test_with_dataset_id(
                                test_run=test_run,
                                dataset_id=data,
                                on_entry_failed=failed_entry_indices.append,
                                on_dataset_finished=mark_all_entries_processed,
                                evaluator_name_to_id_and_pass_fail_criteria_map=evaluator_name_to_id_and_pass_fail_criteria_map,
                            )
                        elif isinstance(data, list):
                            self._run_test_with_local_data(
                                test_run,
                                lambda index: (
                                    data[index] if index < len(data) else None
                                ),
                                failed_entry_indices.append,
                                mark_all_entries_processed,
                                evaluator_name_to_id_and_pass_fail_criteria_map=evaluator_name_to_id_and_pass_fail_criteria_map,
                            )
                        elif isinstance(data, Callable):
                            self._run_test_with_local_data(
                                test_run,
                                data,
                                failed_entry_indices.append,
                                mark_all_entries_processed,
                                evaluator_name_to_id_and_pass_fail_criteria_map=evaluator_name_to_id_and_pass_fail_criteria_map,
                            )
                        else:
                            raise ValueError("Invalid data")

                    self._maxim_apis.mark_test_run_processed(test_run.id)

                    self._config.logger.info(
                        f"You can view your test run here: {self._config.base_url}/workspace/{self._config.in_workspace_id}/testrun/{test_run.id}"
                    )
                    self._config.logger.info(
                        "You can safely quit this session or wait to see the final output in console."
                    )
                except Exception as e:
                    self._maxim_apis.mark_test_run_failed(test_run.id)
                    raise e
                poll_count = 0
                polling_interval = calculate_polling_interval(
                    timeout_in_minutes or 10,
                    is_ai_evaluator_in_use=any(
                        e.type == "AI" for e in evaluator_configs
                    )
                    or False,
                )
                max_iterations = math.ceil(
                    (round(timeout_in_minutes or 10) * 60) / polling_interval
                )
                # Here we will check if we failed to push all entries

                self._config.logger.info("Waiting for test run to complete...")
                self._config.logger.info(
                    f"Polling interval: {polling_interval} seconds"
                )
                status: Optional[TestRunStatus] = None
                sync_check_count = 0
                while True:
                    sync_check_count += 1
                    status = self._maxim_apis.get_test_run_status(test_run.id)
                    if (
                        status is not None
                        and sync_check_count > 5
                        and status.total_entries == 0
                    ):
                        self._config.logger.info(
                            "No entries were pushed to the test run. Exiting..."
                        )
                        break
                    status_dict = status.to_dict()
                    status_line = " | ".join(
                        f"{key}: {value}"
                        for key, value in status_dict.items()
                        if key != "testRunStatus"
                    )
                    box_width = max(50, len(status_line) + 4)
                    header_width = len(
                        f" Test run status: {status.test_run_status.value} "
                    )
                    box_width = max(box_width, header_width + 4)

                    header = (
                        f" Test run status: {status.test_run_status.value} ".center(
                            box_width
                        )
                    )
                    self._config.logger.info("┌" + "─" * box_width + "┐")
                    self._config.logger.info(f"│{header}│")
                    self._config.logger.info("├" + "─" * box_width + "┤")

                    status_line = " | ".join(
                        f"{key}: {value}"
                        for key, value in status_dict.items()
                        if key != "testRunStatus"
                    )
                    self._config.logger.info(f"│ {status_line:<{box_width - 2}} │")
                    self._config.logger.info("└" + "─" * box_width + "┘\n")
                    if poll_count > max_iterations:
                        raise Exception(
                            f"Test run is taking over timeout period ({round(timeout_in_minutes or 10)} minutes) to complete, please check the report on our web portal directly: {self._config.base_url}/workspace/{self._config.in_workspace_id}/testrun/{test_run.id}"
                        )

                    # Test run is failed - we break the loop
                    if (
                        status.test_run_status.value == RunStatus.FAILED.value
                        or status.test_run_status.value == RunStatus.STOPPED.value
                    ):
                        break

                    if (
                        status.test_run_status.value == RunStatus.COMPLETE.value
                        and all_entries_processed.is_set()
                    ):
                        # We will check if we sent all the entries
                        if status.total_entries != 0 and (
                            status.total_entries
                            == status.completed_entries
                            + status.failed_entries
                            + status.stopped_entries
                        ):
                            self._config.logger.info(
                                "All entries processed. Test run completed."
                            )
                            break
                    # Polling again
                    time.sleep(polling_interval)
                    poll_count += 1

                if status.test_run_status.value == RunStatus.FAILED:
                    raise Exception(
                        f"Test run failed, please check the report on our web portal: {self._config.base_url}/workspace/{self._config.in_workspace_id}/testrun/{test_run.id}"
                    )

                if status.test_run_status.value == RunStatus.STOPPED:
                    raise Exception(
                        f"Test run was stopped, please check the report on our web portal: {self._config.base_url}/workspace/{self._config.in_workspace_id}/testrun/{test_run.id}"
                    )

                test_run_result = self._maxim_apis.get_test_run_final_result(
                    test_run.id
                )
                test_run_result.link = self._config.base_url + test_run_result.link
                self._config.logger.info(
                    f'Test run "{name}" completed successfully!🎉 \nView the report here: {test_run_result.link}'
                )
                return RunResult(
                    test_run_result=test_run_result,
                    failed_entry_indices=failed_entry_indices,
                )

            except Exception as e:
                self._config.logger.error("\n\n💥 Error while running test: ", e)

        except Exception as e:
            self._config.logger.error("\n\n💥 Error while running test: ", e)



from dataclasses import dataclass
from typing import Dict, Union


@dataclass
class Metadata():
    """
    This class represents metadata.

    Attributes:
        val: The value of the metadata.
    """

    val: Dict[str, Union[str, int, bool]]



import asyncio
import logging
from typing import Any, AsyncIterator, Awaitable, List, Optional, Union
from uuid import uuid4

from google.genai.chats import AsyncChat
from google.genai.client import AsyncChats, AsyncClient, AsyncModels
from google.genai.types import (
    Content,
    ContentListUnion,
    ContentListUnionDict,
    GenerateContentConfig,
    GenerateContentConfigOrDict,
    GenerateContentResponse,
    PartUnionDict,
)
from typing_extensions import override

from ..logger import (
    Generation,
    GenerationConfig,
    GenerationRequestMessage,
    Logger,
    Trace,
    TraceConfig,
)
from .utils import GeminiUtils


class MaximGeminiAsyncChatSession(AsyncChat):
    """Maxim Gemini async chat session.

    This class represents a maxim gemini async chat session.
    """

    def __init__(
        self,
        chat: AsyncChat,
        logger: Logger,
        trace_id: Optional[str] = None,
        is_local_trace: Optional[bool] = False,
    ):
        """Initialize a maxim gemini async chat session.

        Args:
            chat: The chat.
            logger: The logger.
            trace_id: The trace id.
            is_local_trace: Whether the trace is local.
        """
        super().__init__(
            modules=chat._modules,
            model=chat._model,
            config=chat._config,
            history=chat._curated_history,
        )
        self._chat = chat
        self._logger = logger
        self._trace_id = trace_id
        self._is_local_trace = is_local_trace

    @override
    async def send_message(
        self,
        message: Union[list[PartUnionDict], PartUnionDict],
        generation_name: Optional[str] = None,
    ) -> GenerateContentResponse:
        """Send a message to the chat.

        Args:
            message: The message to send.
            generation_name: The name of the generation.

        Returns:
            GenerateContentResponse: The response from the chat.
        """
        # Without trace_id we can't do anything
        if self._trace_id is None:
            return await super().send_message(message)
        generation: Optional[Generation] = None
        try:
            config = self._chat._config
            messages: List[GenerationRequestMessage] = []
            if config is not None:
                if isinstance(config, GenerateContentConfig):
                    if config.system_instruction is not None:
                        messages.append(
                            GeminiUtils.parse_content(
                                config.system_instruction, "system"
                            )
                        )
                elif isinstance(config, dict):
                    if (
                        instruction := config.get("system_instruction", None)
                    ) is not None:
                        messages.append(
                            GeminiUtils.parse_content(instruction, "system")
                        )
            messages.extend(GeminiUtils.parse_chat_message("user", message))
            gen_config = GenerationConfig(
                id=str(uuid4()),
                model=self._model,
                provider="google",
                name=generation_name,
                model_parameters=GeminiUtils.get_model_params(config),
                messages=messages,
            )
            generation = self._logger.trace_generation(self._trace_id, gen_config)
            # Attaching history as metadata
            if self._curated_history is not None:
                generation.add_metadata({"history": self._curated_history})
        except Exception as e:
            logging.warning(
                f"[MaximSDK][GeminiAsyncClient] Error in generating content: {str(e)}"
            )
        # Actual call will never fail
        response = await super().send_message(message)
        # Actual call ends
        try:
            if generation is not None:
                generation.result(response)
            if self._is_local_trace:
                self._logger.trace_end(self._trace_id)
        except Exception as e:
            logging.warning(
                f"[MaximSDK][GeminiAsyncClient] Error in logging generation: {str(e)}"
            )
        # Returning response
        return response

    @override
    async def send_message_stream(
        self,
        message: Union[list[PartUnionDict], PartUnionDict],
        generation_name: Optional[str] = None,
    ) -> Awaitable[AsyncIterator[GenerateContentResponse]]:
        """Send a message to the chat stream.

        Args:
            message: The message to send.
            generation_name: The name of the generation.

        Returns:
            Awaitable[AsyncIterator[GenerateContentResponse]]: The response from the chat.
        """
        # Without trace_id we can't do anything
        if self._trace_id is None:
            return super().send_message_stream(message)
        generation: Optional[Generation] = None
        try:
            config = self._chat._config
            messages: List[GenerationRequestMessage] = []
            if config is not None:
                if isinstance(config, GenerateContentConfig):
                    if config.system_instruction is not None:
                        messages.append(
                            GeminiUtils.parse_content(
                                config.system_instruction, "system"
                            )
                        )
                elif isinstance(config, dict):
                    if (
                        instruction := config.get("system_instruction", None)
                    ) is not None:
                        messages.append(
                            GeminiUtils.parse_content(instruction, "system")
                        )
            messages.extend(GeminiUtils.parse_chat_message("user", message))
            gen_config = GenerationConfig(
                id=str(uuid4()),
                model=self._model,
                provider="google",
                name=generation_name,
                model_parameters=GeminiUtils.get_model_params(config),
                messages=messages,
            )
            generation = self._logger.trace_generation(self._trace_id, gen_config)
            # Attaching history as metadata
            if self._curated_history is not None:
                generation.add_metadata({"history": self._curated_history})
        except Exception as e:
            logging.warning(
                f"[MaximSDK][GeminiAsyncChatSession] Error in generating content: {str(e)}"
            )
        # Actual call will never fail
        response_awaitable = super().send_message_stream(message)
        # Actual call ends

        # Handle logging asynchronously without consuming the coroutine
        async def handle_logging():
            try:
                if generation is not None:
                    # Create a separate task to consume the stream for logging
                    async def consume_for_logging():
                        async for response in await response_awaitable:
                            if generation is not None:
                                generation.result(response)

                    asyncio.create_task(consume_for_logging())
                if self._is_local_trace:
                    self._logger.trace_end(self._trace_id)
            except Exception as e:
                logging.warning(
                    f"[MaximSDK][GeminiAsyncChatSession] Error in logging generation: {str(e)}"
                )

        # Start logging task without awaiting
        asyncio.create_task(handle_logging())

        # Return the original coroutine
        return response_awaitable

    def __getattr__(self, name: str) -> Any:
        """Get an attribute from the chat.

        Args:
            name: The name of the attribute.

        Returns:
            Any: The attribute.
        """
        result = getattr(self._chats, name)
        return result

    def __setattr__(self, name: str, value: Any) -> None:
        """Set an attribute on the chat.

        Args:
            name: The name of the attribute.
            value: The value of the attribute.
        """
        if name == "_chat":
            super().__setattr__(name, value)
        else:
            setattr(self._chats, name, value)

    def end_trace(self):
        """End the trace.

        This method ends the trace if it is local and the trace id is not None.
        """
        if self._trace_id is not None and self._is_local_trace:
            self._logger.trace_end(self._trace_id)
            self._trace_id = None


class MaximGeminiAsyncChats(AsyncChats):
    """Maxim Gemini async chats.

    This class represents a maxim gemini async chats.
    """

    def __init__(self, chats: AsyncChats, logger: Logger):
        """Initialize a maxim gemini async chats."""
        self._chats = chats
        self._logger = logger
        self._trace_id = None
        self._is_local_trace = False

    @override
    def create(
        self,
        *,
        model: str,
        config: GenerateContentConfigOrDict = None,  # type: ignore
        history: Optional[list[Content]] = None,
        trace_id: Optional[str] = None,
    ) -> AsyncChat:
        """Create a chat session.

        Args:
            model: The model to use.
            config: The config to use.
            history: The history to use.
            trace_id: The trace id.

        Returns:
            AsyncChat: The chat session.
        """
        self._is_local_trace = trace_id is None
        self._trace_id = trace_id or str(uuid4())
        # we start generation here and send it back to chat session
        # every round trip of chat session will be logged in a separate trace
        chat_session = self._chats.create(model=model, config=config, history=history)
        maxim_chat_session = MaximGeminiAsyncChatSession(
            chat=chat_session,
            logger=self._logger,
            trace_id=self._trace_id,
            is_local_trace=self._is_local_trace,
        )
        return maxim_chat_session

    def __getattr__(self, name: str) -> Any:
        """Get an attribute from the chats.

        Args:
            name: The name of the attribute.

        Returns:
            Any: The attribute.
        """
        result = getattr(self._chats, name)
        return result

    def __setattr__(self, name: str, value: Any) -> None:
        """Set an attribute on the chats.

        Args:
            name: The name of the attribute.
            value: The value of the attribute.
        """
        if name == "_chats":
            super().__setattr__(name, value)
        else:
            setattr(self._chats, name, value)


class MaximGeminiAsyncModels(AsyncModels):
    """Maxim Gemini async models.

    This class represents a maxim gemini async models.
    """

    def __init__(self, models: AsyncModels, logger: Logger):
        """Initialize a maxim gemini async models.

        Args:
            models: The models.
            logger: The logger.
        """
        self._models = models
        self._logger = logger

    @override
    async def generate_content_stream(
        self,
        *,
        model: str,
        contents: Union[ContentListUnion, ContentListUnionDict],
        config: Optional[GenerateContentConfigOrDict] = None,
        trace_id: Optional[str] = None,
        generation_name: Optional[str] = None,
    ) -> Awaitable[AsyncIterator[GenerateContentResponse]]:
        """Generate content stream.

        Args:
            model: The model to use.
            contents: The contents to use.
            config: The config to use.
            trace_id: The trace id.
            generation_name: The generation name.

        Returns:
            Awaitable[AsyncIterator[GenerateContentResponse]]: The response from the models.
        """
        is_local_trace = trace_id is None
        final_trace_id = trace_id or str(uuid4())
        generation: Optional[Generation] = None
        trace: Optional[Trace] = None
        try:
            trace = self._logger.trace(TraceConfig(id=final_trace_id))
            # Checking if there is a system prompt
            messages: List[GenerationRequestMessage] = []
            if config is not None:
                if isinstance(config, GenerateContentConfig):
                    if config.system_instruction is not None:
                        messages.append(
                            GeminiUtils.parse_content(
                                config.system_instruction, "system"
                            )
                        )
                elif isinstance(config, dict):
                    if (
                        instruction := config.get("system_instruction", None)
                    ) is not None:
                        messages.append(
                            GeminiUtils.parse_content(instruction, "system")
                        )
            # Adding contents back
            if contents is not None:
                messages.extend(GeminiUtils.parse_messages(contents))

            gen_config = GenerationConfig(
                id=str(uuid4()),
                model=model,
                name=generation_name,
                provider="google",
                model_parameters=GeminiUtils.get_model_params(config),
                messages=messages,
            )
            generation = trace.generation(gen_config)
        except Exception as e:
            logging.warning(
                f"[MaximSDK][GeminiAsyncModels] Error in generating content: {str(e)}"
            )
        # Actual call will never fail
        chunks_awaitable = super().generate_content_stream(
            model=model, contents=contents, config=config
        )
        # Actual call ends

        # Handle logging without consuming the original coroutine
        def handle_completion(task):
            try:
                if generation is not None:
                    result = task.result()
                    generation.result(result)
                if is_local_trace:
                    if trace is not None:
                        # For streams, we'll log what we can get from the result
                        result = task.result()
                        if result is not None:
                            # Note: This is a simplified approach - in practice you might want
                            # to handle streaming differently for comprehensive logging
                            trace.set_output("Stream response received")
                        trace.end()
            except Exception as e:
                logging.warning(
                    f"[MaximSDK][GeminiAsyncModels] Error in logging generation: {str(e)}"
                )

        # Create task and add completion callback
        task = asyncio.create_task(chunks_awaitable)
        task.add_done_callback(handle_completion)

        # Return the task
        return task

    @override
    async def generate_content(
        self,
        *,
        model: str,
        contents: Union[ContentListUnion, ContentListUnionDict],
        config: Optional[GenerateContentConfigOrDict] = None,
        trace_id: Optional[str] = None,
        generation_name: Optional[str] = None,
    ) -> GenerateContentResponse:
        """Generate content.

        Args:
            model: The model to use.
            contents: The contents to use.
            config: The config to use.
            trace_id: The trace id.
            generation_name: The generation name.

        Returns:
            GenerateContentResponse: The response from the models.
        """
        is_local_trace = trace_id is None
        final_trace_id = trace_id or str(uuid4())
        generation: Optional[Generation] = None
        trace: Optional[Trace] = None
        try:
            trace = self._logger.trace(TraceConfig(id=final_trace_id))
            # Checking if there is a system prompt
            messages: List[GenerationRequestMessage] = []
            if config is not None:
                if isinstance(config, GenerateContentConfig):
                    if config.system_instruction is not None:
                        messages.append(
                            GeminiUtils.parse_content(
                                config.system_instruction, "system"
                            )
                        )
                elif isinstance(config, dict):
                    if (
                        instruction := config.get("system_instruction", None)
                    ) is not None:
                        messages.append(
                            GeminiUtils.parse_content(instruction, "system")
                        )
            # Adding contents back
            if contents is not None:
                messages.extend(GeminiUtils.parse_messages(contents))

            gen_config = GenerationConfig(
                id=str(uuid4()),
                model=model,
                provider="google",
                name=generation_name,
                model_parameters=GeminiUtils.get_model_params(config),
                messages=messages,
            )
            generation = trace.generation(gen_config)
        except Exception as e:
            logging.warning(
                f"[MaximSDK][GeminiAsyncModels] Error in generating content: {str(e)}"
            )

        # Actual call will never fail
        response = await self._models.generate_content(
            model=model, contents=contents, config=config
        )
        # Actual call ends
        try:
            if generation is not None:
                generation.result(response)
            if is_local_trace:
                if trace is not None:
                    if response is not None:
                        trace.set_output(response.text or "")
                    trace.end()
        except Exception as e:
            logging.warning(
                f"[MaximSDK][GeminiAsyncModels] Error in logging generation: {str(e)}"
            )
        # Actual response
        return response

    def __getattr__(self, name: str) -> Any:
        """Get an attribute from the models.

        Args:
            name: The name of the attribute.

        Returns:
            Any: The attribute.
        """
        return getattr(self._models, name)

    def __setattr__(self, name: str, value: Any) -> None:
        """Set an attribute on the models.

        Args:
            name: The name of the attribute.
            value: The value of the attribute.
        """
        if name == "_models":
            super().__setattr__(name, value)
        else:
            setattr(self._models, name, value)


class MaximGeminiAsyncClient(AsyncClient):
    """Maxim Gemini async client.

    This class represents a maxim gemini async client.
    """

    def __init__(self, client: AsyncClient, logger: Logger):
        """Initialize a maxim gemini async client.

        Args:
            client: The client.
            logger: The logger.
        """
        self._client = client
        self._logger = logger
        self._w_models = MaximGeminiAsyncModels(client.models, logger)
        self._w_chats = MaximGeminiAsyncChats(client.chats, logger)

    def __getattr__(self, name: str) -> Any:
        """Get an attribute from the client.

        Args:
            name: The name of the attribute.

        Returns:
            Any: The attribute.
        """
        if name == "_models":
            return self._w_models
        elif name == "_chats":
            return self._w_chats
        result = getattr(self._client, name)
        return result

    def __setattr__(self, name: str, value: Any) -> None:
        """Set an attribute on the client.

        Args:
            name: The name of the attribute.
            value: The value of the attribute.
        """
        if name == "_client":
            super().__setattr__(name, value)
        else:
            setattr(self._client, name, value)
